import numpy as np
from numpy import nanstd, nanmean
#from matplotlib.mlab import prepca
from os.path import join as osjoin
from os import listdir as listdir
import pickle

from baseAnalysis import BaseAnalysis
from utils import GPInterpolater
from mDInterpret import getID, loadMetaData
from scipy.stats import linregress
from sklearn import decomposition
from utils import lpfiltbutter,fillNans,FiltFiltFilter

class BeltVelocity(BaseAnalysis):
    defaultTargets = ["*tVelocity","*tMarkerSet"]
    defaultIDExt = "bvel"
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        '''
        Constructor, loads suitable meta data after the parent class has had 
        a chance to set any defaults.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch,        
        '''
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self.source = self.parent.resolve(self.targets[0])
        self.markers = self.parent.resolve(self.targets[1])
        self.sampleRate = self.source.sampleRate
        self.legs = [leg for leg in self.markers["Appendages"]]
        self.feet = [self.markers[leg+"Markers"][0] for leg in self.markers["Appendages"]]
        loadMetaData("beltDirection",self,self.metaData,[1,0,0])
        loadMetaData("thold",self,self.metaData,0.9995)


    def _getTreadmillBeltSpeed(self):
        '''
        Computes the treadmill belt speed from foot markers.
        Only works if trialType == 'treadmill'.
        At present works for horses, and specifically uses right fore hoof markers.
        Computes a normalized (unit magnitude) direction vector of the hoof velocity,
        and then thresholds the dot product of this against treadmill belt direction
        to determine time points when the hoof is moving in the same direction
        as the treadmill belt (irrespective of velocity magnitude). Then interpolates
        these segments of speed and low pass filters them to come up with a belt speed
        for each time point.
        '''
        if self.metaData["verbosity"]>0:
            print "beltVelocity object iD: %s computing belt speed." % self.iD
        for foot,leg in zip(self.feet,self.legs):
            # Get foot marker velocity
            m = self.source[foot+"_velocity"]
            # Is foot aligned? Fraction along belt is vdir, velocity is 
            # vmag, store aligned points in in va
            vdir = np.sum(np.array(self.beltDirection)*m.transpose(),1)/np.sqrt(np.sum(m**2.0,0))
            vmag = np.sum(np.array(self.beltDirection)*m.transpose(),1)
            va = find( vdir > self.thold ) # This may be empty!! Either inspect and lower threshold, or allow legs to be all nan. Former easier first instance 
            
            # Compute belt speed, start off with an array of nans of right length
            bs = np.nan*np.ones(m.shape[1])
            if len(va) > 0:
                bs[va] = vmag[va]
                bs[0:va[0]] = vmag[va[0]]
                bs[va[-1]:-1] = vmag[va[-1]]
                bsi = fillNans(bs)
            else:
                bsi=np.copy(bs)
            self.data[leg+u"treadmillbelt_speed"]=bsi#lpfiltbutter(bsi,self.sampleRate,0.5)
            self.data[leg+u"treadmillbelt_hoofveldotbeltdir"]=vdir
            self.data[leg+u"treadmillbelt_hoofonindices"]=va
            self.data[leg+u"treadmillbelt_validspeeds"]=np.copy(bs)
            self.data[leg+u"treadmillbelt_validspeedsinterp"]=bsi
        ff = FiltFiltFilter(self.sampleRate,1.0,order=2)
        self.data["BeltSpeed"] = ff(fillNans(nanmean([self.data[leg+'treadmillbelt_validspeeds'] for leg in self.legs],0)))
        
    def keys(self):
        '''
        Computes belt speeds and related
        '''
        computedValues = [
                          u"treadmillbelt_speed", u"treadmillbelt_hoofveldotbeltdir",
                          u"treadmillbelt_hoofonindices", u"treadmillbelt_validspeeds",
                          u"treadmillbelt_validspeedsinterp"
                         ]
        return(["BeltSpeed"]+[leg+c for leg in self.legs for c in computedValues])

    def has_key(self, k):
        '''
        Is the key k indexable? Just check is k is in self.keys()
        @param k: key
        @type k: any immutable
        '''
        return(k in self.keys())


    def __getitem__(self,k):
        '''
        Returns the value associated with key k.
        @param k: key
        @type k: any immutable
        '''
        # Mangle the key first
        if not self.data.has_key(k):
            self._getTreadmillBeltSpeed()
        return self.data[k]
        raise KeyError("beltVelocity analyses only contain information on treadmill belt speed! Invalid key %s",k)

        