'''
Author: Simon Wilshin
Created: 11 Oct 2011

Analyses for finding duty factor from different data sets. At present this just looks at 
motion capture data and finds when the foot is above some fraction of the stride.
'''

import numpy as np
import pylab as pl

from baseAnalysis import BaseAnalysis
from mDInterpret import getID
from mDInterpret import loadMetaData

class DutyFactor(BaseAnalysis):
    '''
    A plugin for finding the duty factor of strides cut by some stride cutter 
    from some motion capture data. Requires knowledge of a marker set so that 
    it can determine which markers are the legs.
    @param footOnFrac: Fraction of height that a foot is above the ground which
    should be considered on the ground (0.1 = below 10% of max height foot on 
    ground)
    @type footOnFrac: float
    '''
    defaultTargets = ["*tMarkerSet","*tMotionCaptureData","*tPhasicStrideCutter"]
    defaultIDExt = "df"
    def __init__(self,workingDirectory,analysisDirectory, metaData, parent):
        '''
        Constructor.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch, 
        '''
        BaseAnalysis.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self.markers = self.parent.resolve(self.targets[0])
        self.source = self.parent.resolve(self.targets[1])
        self.pSC = self.parent.resolve(self.targets[2])
        self.sampleRate = self.source.sampleRate
        loadMetaData("footOnFrac",self,self.metaData,0.1)
        loadMetaData("footHeightThresh",self,self.metaData,5.0)

    def plotDutyFactor(self,k):
        '''
        Plot the duty factor of leg k, or rather plot how it was obtained.
        @param k: key
        @type k: any immutable
        '''
        pl.figure()
        foot = self.source[self.markers[k+"Markers"][0]][2]
        pl.plot(self.source.time, self._getFootState(k))
        pl.plot(self.source.time, self.source[self.markers[k+"Markers"][0]][2])
        for i,j in self.pSC["Body_Strides"]:
            pl.axvline(self.source.time[i])
        
    def _getFootState(self,k):
        foot = self.source[self.markers[k+"Markers"][0]][2]
        fstate = np.zeros_like(foot)
        for i,j in self.pSC["Body_Strides"]:
            #fstate[i:j] = np.array((foot[i:j]-foot[i:j].min())/(foot[i:j].max()-foot[i:j].min())<self.footOnFrac)
            fstate[i:j] = np.array((foot[i:j]-foot[i:j].min())<self.footHeightThresh)
        return fstate
        
    def has_key(self,k):
        '''
        Is the key k indexable? (Only true is k identifies a 
        DutyFactor signal, or a DutyFactorList. Just check is 
        k is in self.keys()
        @param k: key
        @type k: any immutable
        '''
        return(k in self.keys())

    def keys(self):
        '''
        Can compute a continuous DutyFactor and a DutyFactorList 
        for each appendage, this member will report a list of these 
        possible indices.
        '''
        return(
               ["_".join([k,"DutyFactor"]) for k in self.markers["Appendages"]]+
               ["_".join([k,"DutyFactorList"]) for k in self.markers["Appendages"]]
              )

    def _getDutyFactor(self, k):
        '''
        Compute the DutyFactor for appendage k.
        @param k: A appendage
        @type k: string
        '''
        self.data[k+"_DutyFactor"] = np.zeros_like(
                                                   self.source[
                                                    self.markers[
                                                     k+"Markers"
                                                    ][0]
                                                   ][2]
                                                  ) # Going to fill this with duty factors
        self.data[k+"_DutyFactorList"] = list()
        # Average over each stride from the body phase
        fstate = self._getFootState(k)
        for i,j in self.pSC["Body_Strides"]:
            self.data[k+"_DutyFactor"][i:j] = np.sum(fstate[i:j])/float(j-i)
            self.data[k+"_DutyFactorList"].append(np.sum(fstate[i:j])/float(j-i))
            
    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Because certain quantities are computationally expensive to determine
        this method now checks to see if these have been calculated when 
        accessed. If they have not it calls the method to compute them. This 
        avoids the need to determine the at start up.
        Returns: The data associated with k.
        @param k: key
        @type k: any immutable
        '''
        # Mangle the key first
        nk=self.mangleKey(k)
        if (type(nk)==type(str()) or type(nk)==type(u"")):
            if not self.data.has_key(nk):
                if nk[-11:]=="_DutyFactor":
                    self._getDutyFactor(nk[:-11])
                if nk[-15:]=="_DutyFactorList":
                    self._getDutyFactor(nk[:-15])
        return(self.data[k])

