'''
Created on 14 Feb 2011

@author: swilshin
'''

from markerSet import MarkerSetInterface
from markerSet import LabelException
from mDInterpret import getID

class FishMarkerSet(MarkerSetInterface):
    '''
    A direct implementation of the fish marker set provided by Dr. Wiktorowicz-Conroy.
    A seperate marker set class is needed for this just because fish are so completely 
    different to terrestrial animals that terrestrial animal anatomy does not transfer 
    readily.
    '''
    defaultIDExt = "fms"
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        '''
        Constructor.
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param workingDirectory: The working directory relative to which this objects 
        working directory should be resolved
        @type workingDirectory: string
        @param metaData: Standard metadata dictionary.
        @type metaData: dict
        @param parent: Parent in the experiment tree.
        @type parent: ExperimentBranch, 
        '''
        MarkerSetInterface.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        self.appendageNames = self.appendageMarkers.keys()
        self._checkMarkerClassification()

    def _detectJunkMarkers(self):
        '''
        Find the junk markers, there isn't a while lot 
        that can be done with the ventral cal marker, so 
        this has been labelled junk for now.
        '''
        self.junkMarkers=["ventral_cal"]
    
    def _detectAppendages(self):
        '''
        Since the fish markers are rather unique, this just records 
        the appendages directly rather than going to the trouble to 
        build a regular expression based classifier.
        '''
        self.appendageMarkers={}
        self.appendageMarkers["analFinAnterior"]=["ventral_analantorg","ventral_analanttip"]
        self.appendageMarkers["analFinPosterior"]=["ventral_analpostorg","ventral_analposttip"]
        self.appendageMarkers["caudalFin"]=["ventral_lowcaudorg","ventral_lowcaudtip"]
        self.appendageMarkers["leftPectoralFin"]=["ventral_pecorglead_left","ventral_pectiplead_left"]
        self.appendageMarkers["rightPectoralFin"]=["ventral_pecorglead_right","ventral_pectiplead_right"]

        
    def _detectBodyMarkers(self):
        '''
        Since the fish markers are rather unique, this just records 
        the body markers directly rather than going to the trouble to 
        build a regular expression based classifier.
        '''
        self.bodyMarkers=["ventral_analantorg","ventral_analpostorg","ventral_lowcaudorg","ventral_pecorglead_left","ventral_pecorglead_right"]
    
    def _detectHeadMarkers(self):
        '''
        Since the fish markers are rather unique, this just records 
        the head marker directly rather than going to the trouble to 
        build a regular expression based classifier.
        '''        
        self.headMarkers=["ventral_jawant"]
    
    def _checkMarkerClassification(self):
        '''
        All markers should be classified. If they are not this is problem and 
        we should throw an exception.
        '''
        allIdentifiedMarkers=self.bodyMarkers+self.headMarkers+self.junkMarkers
        for leg in self.appendageMarkers:
            allIdentifiedMarkers+=self.appendageMarkers[leg]
        if set(self.markerNames)-set(allIdentifiedMarkers)!=set([]):
            raise(LabelException("Some markers were not identified! Unidentified markers: "+str(set(self.markerNames)-set(allIdentifiedMarkers))))
    
    def soundOff(self):
        '''
        Report on the identified markers
        '''
        print "All markers ", self.markerNames
        print "Body Markers ", self.bodyMarkers
        print "Head Markers ", self.headMarkers
        print "Junk Markers ", self.junkMarkers 
        print "Fins ", self.appendageMarkers
        print "Fin Names ", self.getAppendages()
    
def testFishMarkerSet():
    '''
    Load up a test case using the trial type format of the experiment
    branch.
    '''
    from experimentTree import ExperimentBranch
    from os.path import join as osjoin
    from yaml import load as yamlload
    workingDir = osjoin("..","sampleData","fishSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","analysisContainers")
    metaDataFile = osjoin("..",'sampleYAML','trialContainers','fishMarkerSetTest.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    a = ExperimentBranch(workingDir, analysisDir, metaData, None)
    a.tryResolve()
    return(a)

if __name__=='__main__':
    '''
    Runs some tests using the sample data. At the moment 
    just invokes the sample fish data using the test yaml.
    '''
    print "Processing sample fish marker set"
    a = testFishMarkerSet()