'''
Created on 18 Jan 2011

@author: Simon Wilshin, swilshin@rvc.ac.uk

Data objects for manipulating force plate data including a base class (not overly useful) 
and two derived classes, on for weighing data, one for kinematic data.
'''

import csv
import numpy as np
import pylab
from os.path import join as osjoin
from scipy import nanmean
from warnings import warn
from yaml import load as yamlload

import utils
import conventions as conv
from baseData import BaseData
from temporalSignal import TemporalSignal
from weightInterface import WeightInterface

class BaseForcePlateData(BaseData):
    '''
    A base class for all data types corresponding to Force Plate data
    '''
    def __init__(self, workingDirectory, analysisDirectory, metaData):
        BaseData.__init__(self, workingDirectory, analysisDirectory, metaData)
        self.numberOfPlates=metaData["numberOfPlates"]
        self.zeroWidth=metaData["zeroWidth"]
        self.zeroVarianceFactor=float(metaData["zeroVarianceFactor"])
        if metaData.has_key("zeroSearchStep"):
            self.zeroSearchStep=int(metaData["zeroSearchStep"])
        else:
            self.zeroSearchStep=int(self.sampleRate/20.0)      
        if metaData.has_key("zeroPadding"):
            self.zeroPadding=int(metaData["zeroPadding"])
        else:
            self.zeroPadding=int(self.zeroWidth/4.0)
        self.resolventKeys=conv.makeResolventForcePlateKeys(self.numberOfPlates)
        self.baseKeys=conv.makeForcePlateKeys(self.numberOfPlates)
        self._readData()
        self._zeroPlates()

    def _readData(self):
        '''
        Read data from a standard N-plate force plate file. Expect data to be laid out as in 
        a presently undocumented format specification (BUG0001).
        '''
        print "Reading Force Plate Data..."
        # Open the file
        file = open(osjoin(self.workingDirectory, self.fileName),'rb')
        # get a csv reader object
        fileCSV = csv.DictReader(file,delimiter='\t',fieldnames=self.baseKeys)        
        # Make empty lists to be populated by our data
        for k in self.baseKeys:
            self.data[k]=list()
        self.totalSamples=None
        # Read the data one row at a time using the keys just determined above.
        for row in fileCSV:
            for k in row.keys():
                self.data[k].append(float(row[k]))
        # If the total number of samples has not already been set, then set it.
        if not self.totalSamples:
            self.totalSamples=len(self.data[k])
        # Calculate an array containing the time of each sample in seconds since 
        # the start of the signal.
        self.time=np.array(range(0,self.totalSamples))/float(self.sampleRate)
        # Convert the result to a temporal signal with its own attached time
        for k in self.baseKeys:
            self.data[k]=TemporalSignal(self.data[k],self.time,0)
        # Close the file
        file.close()   
        
    def _calculateResolvents(self):
        '''
        Get resolvent (forces in a direction summed over the sensors 
        in the plate) forces on a plate by plate basis.
        '''
        print "Calculating Resolvent Forces..."
        # For every sensor in our plate...
        for j in self.resolventKeys:
            # Since we calculate things cumulatively we start of with a long list of zeros
            self.data[j]=TemporalSignal(np.zeros(self.totalSamples),self.time,0)
            for k in self.baseKeys:
                # If the direction of the sensor matches the direction of the resolvent 
                # we want, add its contribution 
                if k.direction[0]==j.direction[0] and k.plate==j.plate:
                    self.data[j]+=self.data[k]             

    def _calculateTotalForces(self):
        '''
        Calculate the total resolvent forces summed over all of the plates.
        '''
        print "Computing Total Forces by direction..."
        # For every direction in space (there are 3)
        for j in conv.spaceDirections3:
            # Since we calculate things cumulatively we start of with a long list of zeros
            self.data[j]=TemporalSignal(np.zeros(self.totalSamples),self.time,0)
            for k in self.baseKeys:
                # If the direction of this sensor matches the direction of interest
                # add it's contribution
                if k.direction[0]==j[0]:
                    self.data[j]+=self.data[k]
                    
    def _zeroPlates(self):
        '''
        Force plates have a low frequency drifting bias. Here we look for a nice flat 
        region (which should be the zero) on each plate and remove the mean of this. The 
        zero then really should be zero force.
        '''
        print "Zeroing Force Plates..."
        for k in self.baseKeys:
            # Find the variance of the signal over all. We will look for regions 
            # which are some suitably small fraction of this (how small is determined 
            # by the parameter self.zeroVarianceFactor).
            baseVariance=np.var(self.data[k])
            flatRegions=utils.findFlatRegions(self.data[k],
                                              self.zeroWidth,
                                              baseVariance/self.zeroVarianceFactor,
                                              self.zeroSearchStep)
            # If we have found some flat regions then subtract the mean of the longest one.
            if flatRegions!=[]:
                # Get the lengths of the flat regions and corresponding indicies
                ls=[(b-a,a,b) for a,b in flatRegions]
                # Sort them, the longest is now the last entry of this list
                longest=sorted(ls,key=lambda k: k[0])[-1]
                # If the signal is long enough that with the padding we subtract 
                # there is still some signal left then
                # subtract this longest regions mean value from the signal
                if longest[0]>2.0*self.zeroPadding:
                    self.data[k]-=np.mean(self.data[k][(longest[1]+self.zeroPadding):(longest[2]-self.zeroPadding)])
                # Otherwise do the best job you can, remove the mean of the 
                # longest region and print a warning.
                else:
                    print "Warning: Longest zero region of force plate data was too short to add desired padding!"
                    self.data[k]-=np.mean(self.data[k][longest[1]:longest[2]])
            # If there are no flat regions just subtract the mean (usually this means 
            # that nothing was put on the plate
            else:
                print "Warning: There were no sufficiently long flat regions with which to zero the signal from plate "+str(k.plate)+" direction "+str(k.direction)
                self.data[k]-=np.mean(self.data[k])     
                self.data["_".join([str(k.plate),k.direction,"noise"])]=np.std(self.data[k])                    
        
class ForcePlateWeightData(BaseForcePlateData,WeightInterface):
    '''
    Use this when the force plates have been used to weigh an animal. 
    '''
    def makePlots(self):
        '''
        Plot a graph which illustrates the weight calculation by marking horizontal bars
        where the weights were taken from and annotates the graph.
        '''
        print "Plotting weight signal"
        # Set figure size
        pylab.gcf().set_size_inches( conv.pylabImageSizeX, conv.pylabImageSizeY)
        pylab.clf()
        # Plot
        pylab.plot(self["Z"].getTime(),self["Z"],'r')
        # Labels and titles
        pylab.xlabel("time (s)")
        pylab.ylabel("Weight (N)")
        pylab.title("Weight against time for the weight trial, Weight: " + str(self.weight) + " N, mass: " + str(self.mass) + " kg")
        # Line marking the region used to weigh the animal, include one below the axis in case zeroing has been done poorly
        pylab.axhline(y=self.weight,color='k')
        pylab.axhline(y=-self.weight,color='k')
        # Save figure
        pylab.savefig(osjoin(self.analysisDirectory, "_weighed.".join([self.fileName[:-4],"png"])))
        # Clear
        pylab.clf()
            
    def _determineWeight(self):
        '''
        Compute the weight of our animal by finding the two longest periods where the 
        variation of the z-component of the force is not varying and subtract them.
        '''
        weightEstimates=[]
        v0=np.var(self["Z"])
        # Get the two flat regions, two long ones should exist whose difference is 
        # vaguely correct will give us the weight
        flatRegions=self["Z"].findFlatRegions(self.flatW,v0/self.flatVarFac,self.flatStep)
        if len(flatRegions)>=2:
            ls=[b-a for a,b in flatRegions]
            ls=zip(*sorted(zip(ls,range(0,len(ls))),key=lambda k: k[0]))[1]
            weightEstimates.append(np.abs(np.mean(self["Z"][(flatRegions[ls[-2]][0]+self.padding):(flatRegions[ls[-2]][1]-self.padding)])))
            return(min((abs(self.weightEstimate - w), w) for w in weightEstimates)[1])
        # (!!!) Should test here for extreme outliers at some point
        else:
            warn("Weight estimation from force plate data has not found a good estimate of weight. Estimated weight used instead")
            return(self.weightEstimate)

    
    def __init__(self, workingDirectory, analysisDirectory, metaData):
        '''
        Constructor. First we assemble the BaseData part of the object, then 
        process the meta data. Then compute our animals weight.
        '''
        BaseForcePlateData.__init__(self, workingDirectory, analysisDirectory, metaData)
        # Make our base object that handles directories, etc.
        if metaData.has_key("flatW"):
            self.flatW = metaData["flatW"]
        else:
            self.flatW = 0.2*self.sampleRate # Width a region has to be to be called flat
        if metaData.has_key("flatVarFac"):
            self.flatVarFac = float(metaData["flatVarFac"])
        else:
            self.flatVarFac = 30 # What fraction of the variance of the signal
        if metaData.has_key("flatStep"):
            self.flatStep = metaData["flatStep"]
        else:
            self.flatStep = 0.010 * self.sampleRate # How fine grained the search for flat regions should be
        if metaData.has_key("padding"):
            self.padding=metaData["padding"]
        else:
            self.padding = 0.029 * self.sampleRate # How much to remove from the start and end of a flat region (to ensure it doesn't have transients)
        if metaData.has_key("weightEstimate"):
            self.weightEstimate = metaData["weightEstimate"]
        else:
            self.weightEstimate = 240 # A reasonable first guess 
        self._calculateTotalForces()
        WeightInterface.__init__(self,self._determineWeight())

class ForcePlateData(BaseForcePlateData):
    '''
    Manipulates force plate data. Calculates things like resolvent forces, total 
    forces, order of the force plates, peak detection, peak refinement, stride 
    frequency estimation and detects if a force signal is split over multiple plates.
    '''

    def _doPlot(self,keys,colourScheme,title,id):
        '''
        Saves a plot of forces in data with the keys keys, using the colour scheme 
        specified by colourScheme (should be a list of line styles compatible with 
        pylab). The plot is given the title title and the name of the save file will 
        include the text specified in id.
        The peaks and troughs are added to these graphs.   
        '''
        # Set figure size
        pylab.gcf().set_size_inches( conv.pylabImageSizeX, conv.pylabImageSizeY)
        # Plot
        for k in keys:
            pylab.plot(self.data[k].getTime(),self.data[k],colourScheme[k])
        # Labels and titles
        pylab.xlabel("time (s)")
        pylab.ylabel("force (N)")
        pylab.title(title)
        # Lines marking peaks and troughs
        [pylab.axvline(x=p,color='k') for p in self.data["peaks"]]
        [pylab.axvline(x=p,color='c') for p in self.data["troughs"]]
        # Save figure
        pylab.savefig(osjoin(self.analysisDirectory, ".".join(["_".join([self.fileName,id]),"png"])))
        # Clear
        pylab.clf()

    def makePlots(self):
        '''
        Makes a set of standard plots. Currently plots raw forces, plate specific 
        resolvent forces,  and total forces. It also plots the location of plate 
        specific peaks along with total force and the weighted center of 
        mass displacement (without initial conditions).
        '''
        print "Plotting Force Plate Data..."
        # Raw data
        self._doPlot(self.baseKeys,conv.makeForcePlateColourScheme(self.numberOfPlates), "forces against time","raw")
        # Resolvents per plate
        self._doPlot(self.resolventKeys,conv.makeResolventForcePlateColourScheme(self.numberOfPlates), "Total force against time by plate","total_by_plate")
        # Total resolvent across plates
        self._doPlot(conv.spaceDirections3,conv.makeSpaceColourScheme(), "Total force against time","total")
        # Plot of total resolvent force summed across plates with plate specific 
        # peaks and troughs.
        pylab.gcf().set_size_inches( conv.pylabImageSizeX, conv.pylabImageSizeY)
        for k in conv.spaceDirections3:
            pylab.plot(self[k].getTime(),self[k],conv.makeSpaceColourScheme()[k])
        pylab.xlabel("time (s)")
        pylab.ylabel("force (N)")
        for k in self.resolventKeys:
            if k.direction=="Z":
                [pylab.axvline(x=p,color=conv.plateColours[k.plate]) for p in self.data["_".join([str(k.plate),"peaks"])]]
        pylab.savefig(osjoin(self.analysisDirectory, ".".join(["_".join([self.fileName[:-4],"individualPlatePeaks"]),"png"])))
        pylab.clf()
        # Plot the weighted center of mass displacement with unknown initial conditions
        pylab.gcf().set_size_inches( conv.pylabImageSizeX, conv.pylabImageSizeY)
        for k in conv.spaceDirections3:
            pylab.plot(self["_".join(["Fdtdt",k])].getTime(),self["_".join(["Fdtdt",k])],conv.makeSpaceColourScheme()[k],label=k)
        pylab.xlabel("time (s)")
        pylab.ylabel("Weighted CoM displacement (kgm)")
        pylab.title("Mass weighted CoM displacement against time")
        pylab.legend()
        pylab.savefig(osjoin(self.analysisDirectory, ".".join(["_".join([self.fileName[:-4],"weightedCoMDisplacement"]),"png"])))
        
    def _selectPeaks(self):
        '''
        Finds the peaks in the total Z componenet of the force
        '''
        print "Finding Interesting Peaks..."
        # Filter the z component of total force and find peaks
        strideSignal=self.data["Z"]
        interpolatedStrideSignal=strideSignal.nanInterpolation(left=nanmean(strideSignal),right=nanmean(strideSignal))
        highFrequencyCutOff = self.maxFrequencyOfInterestMultiplier * self.inferredFrequency
        filteredStrideSignal=interpolatedStrideSignal.filter(self.sampleRate,highFrequencyCutOff,self.filterOrder,'lowpass')
        peaks=utils.findPeaks(filteredStrideSignal,self.absurdForcePeak,self.maxPeakRangeFraction,self.maxForceVariance)
        self.data["peaks"]=np.array([self.time[k] for k in zip(*peaks)[0]])
        self.data["peakIndicies"]=np.array([k for k in zip(*peaks)[0]])

    def _refinePeaks(self):
        '''
        Refines the peaks previously found by _selectPeaks method. To 
        do this at present this is done by just finding the highest 
        point between two troughs found by _findTroughs. This 
        is a local peak finding method and as such will be subject to 
        noise. This could be corrected by spline fitting or using a 
        filter which does not shift the peaks (not a butterworth!) 
        prior to peak detection.
        '''
        #mTroughs=[0]+list(self.data["troughIndicies"])+[len(self.data["Z"])]
        mTroughs=list(self.data["troughIndicies"])
        peaks=[s+self.data["Z"][s:e].argmax() for s,e in zip(mTroughs[:-1],mTroughs[1:])]
        self.data["peaks"]=np.array([self.time[k] for k in peaks])
        self.data["peakIndicies"]=np.array([k for k in peaks])        

    def _findTroughs(self):
        '''
        Find the troughs between peaks by finding the minima which lies 
        between each sucessive pair of peaks. This is a local method 
        and is subject to noise. As such it should be replaced at some 
        point by a non-local peak detection method (such as using 
        spline fitting or filtering (not a butterworth!) prior to 
        peak detection.
        '''
        print "Finding troughs..."
        # Add phantom peaks to the start and end of the peak data so 
        # as to find 'troughs' at either end of the spectrum and enclose 
        # all peaks
        mPeaks=[0]+list(self.data["peakIndicies"])+[len(self.data["Z"])-1]
        # Find the minimum value between each pair of peaks and call this 
        # a trough.
        troughs=[s+self.data["Z"][s:e].argmin() for s,e in zip(mPeaks[:-1],mPeaks[1:])]
        self.data["troughs"]=np.array([self.time[k] for k in troughs])
        self.data["troughIndicies"]=np.array([k for k in troughs])

    def _findTotalDoubleIntegratedForce(self):
        '''
        Integrates the force twice, this calculation will be necessary for 
        finding the center of mass position.
        '''
        print "Calculating double integrated Total Force..."
        for d in conv.spaceDirections3:
            self.data["_".join(["Fdtdt",d])]=TemporalSignal(
                                                    utils.wrappedIntegrate(utils.wrappedIntegrate(self.data[d],self.data[d].getTime()),self.data[d].getTime()),
                                                    self.data[d].getTime(),
                                                    0
                                                    )

    def _plateSpecificPeaksAndTroughs(self):
        '''
        Here we find peaks and troughs in the Z component specific to 
        each plate. At some point this and the _findPeaks, and related 
        functions should be rolled into a single utility since there is 
        some duplicate code. In both cases peaks are found using the 
        utils function find peaks which factorises some of the 
        duplication, but I (SW) think more could be done. 
        '''
        print "Finding peaks and troughs in specific plate signals..."
        # For each of the resolvent keys look at the Z direction, filter 
        # and find peaks using the utils.findPeaks method.
        for k in self.resolventKeys:
            if k.direction=="Z":
                strideSignal=self.data[k]
                interpolatedStrideSignal=strideSignal.nanInterpolation(left=nanmean(strideSignal),right=nanmean(strideSignal))
                highFrequencyCutOff = self.maxFrequencyOfInterestMultiplier * self.inferredFrequency
                filteredStrideSignal=interpolatedStrideSignal.filter(self.sampleRate,highFrequencyCutOff,self.filterOrder,'lowpass')
                peaks=utils.findPeaks(filteredStrideSignal,self.absurdForcePeak,self.maxPeakRangeFraction,self.plateSpecificMaxForceVariance)
                self.data["_".join([str(k.plate),"peaks"])]=np.array([self.time[q] for q in zip(*peaks)[0]])
                self.data["_".join([str(k.plate),"peakIndicies"])]=np.array([q for q in zip(*peaks)[0]])
                # To find troughs add phantom peaks so one find troughs at the etnd of 
                # the singal, then find either threshold values or minimal values.
                mPeaks=[0]+list(self.data["_".join([str(k.plate),"peakIndicies"])])+[len(strideSignal)-1]
                troughs=[s+strideSignal[s:e].argmin() for s,e in zip(mPeaks[:-1],mPeaks[1:])]
                # First peak is special because we are certain there is no 
                # peak to the left.
                prevIdx, nextIdx = mPeaks[0],mPeaks[1]
                # Look at index prevIdx+tau (tau increases with each loop) 
                tau=0
                converged=False
                while not converged:
                    # Check if we have risen above the threshhold value.
                    if strideSignal[prevIdx+tau]/strideSignal[nextIdx]>self.peakThreshold:
                        troughs[0]=prevIdx+tau
                        converged=True
                    # If we haven't found a region above the threshold just take the 
                    # minimum value.
                    elif prevIdx+tau>=nextIdx:
                        troughs[0]=prevIdx+strideSignal[prevIdx:nextIdx].argmin()
                    # If we haven't converged and we haven't ran out of space
                    # in which to search, then increment tau (where we are looking)
                    else:
                        tau+=1
                # Last peak is special because we know there is no peak to the right of it
                prevIdx, nextIdx = mPeaks[-2], mPeaks[-1]
                # Look at index prevIdx+tau (tau increases with each loop)
                tau=0
                converged=False
                while not converged:
                    # Check if we have fallen below the threshold value
                    if strideSignal[prevIdx+tau]/strideSignal[prevIdx]<self.peakThreshold:
                        troughs[-1]=prevIdx+tau
                        converged=True
                    # If we have run out of space in which to look just take the minimum value
                    elif prevIdx+tau>=nextIdx:
                        troughs[-1]=prevIdx+strideSignal[prevIdx:nextIdx].argmin()
                    # If we haven't converged and haven't ran out of space 
                    # in which to look then increment tau (where we are looking)
                    else:
                        tau+=1    
                # Store these peaks and troughs.   
                self.data["_".join([str(k.plate),"troughs"])]=np.array([self.time[q] for q in troughs])
                self.data["_".join([str(k.plate),"troughIndicies"])]=np.array([q for q in troughs])
                mTroughs=list(self.data["_".join([str(k.plate),"troughIndicies"])])
                peaks=[s+strideSignal[s:e].argmax() for s,e in zip(mTroughs[:-1],mTroughs[1:])]
                self.data["_".join([str(k.plate),"peaks"])]=np.array([self.time[q] for q in peaks])
                self.data["_".join([str(k.plate),"peakIndicies"])]=np.array([q for q in peaks])
                 
    def _estimateStrideFrequency(self):
        '''
        Estimate the stride frequency by finding the peak of the power spectrum of the signal minus 
        the DC part. Crude but effective.
        '''
        print "Estimating Stride Frequency..."
        strideSignal=self.data["Z"]-np.mean(self.data["Z"])
        self.inferredFrequency=utils.inferPeakFrequency(strideSignal, self.sampleRate,self.minFrequencyOfInterestMultiplier*self.strideFrequencyEstimate,self.maxFrequencyOfInterestMultiplier*self.strideFrequencyEstimate,self.filterOrder)
        print "Inferred stride frequency of ", self.inferredFrequency
        
    def _classifySignalType(self):
        '''
        Check if the signal is split or not. A split signal is characterised by two plates 
        containing a significant fraction of the applied force to the plates at the same time.
        '''
        self.data["splitSignal"]=TemporalSignal([conv.split[sc] for sc in 
                        (np.max(np.abs([self[k] for k in self.resolventKeys if k.direction=="Z"]),0)
                         /
                         np.sum(np.abs([self[k] for k in self.resolventKeys if k.direction=="Z"]),0)
                        )<self.splitFraction],
                        self.time,
                        0
                        )

    def _orderPlates(self):
        '''
        Put the plates in order depending on when peaks are in the signal 
        The first plate has the first peak, and so on.
        '''
        print "Ordering the force plates..."
        # Pair the plates k (looking only at Z direction) with their earliest peak locations. 
        plateOrder=[[k,min(self.data["_".join([str(k.plate),"peaks"])])]
                    for k in self.resolventKeys
                    if k.direction=="Z"
                   ]
        # Order this list according to where these peaks are
        plateOrder=sorted(plateOrder,key=lambda p: p[1])
        # Get the resulting ordering of the plates
        self.plateOrder=zip(*plateOrder)[0]
         
         
    def firstPlateSecondPlate(self,p0,p1):
        '''
        Tells you if p0 comes before p1 in the plate ordering. If it does 
        this returns True.
        '''
        # If either of the plate identifies are tuples convert them
        if type(p0)==tuple:
            p0=conv.ForcePlateIdentifierTuple(plate=p0[0],direction=p0[1])
        if type(p1)==tuple:
            p1=conv.ForcePlateIdentifierTuple(plate=p1[0],direction=p1[1])
        # Return true if p0 is before p1.
        return self.plateOrder.index(p0)<self.plateOrder.index(p1) 

    def __init__(self, workingDirectory, analysisDirectory, metaData):
        '''
        Constructor. First we assemble the BaseData part of the object, then 
        process the meta data. Finally we run each of the standard methods 
        to populate the object with the information we care about.
        '''
        # Make our base object that handles directories, etc.
        BaseForcePlateData.__init__(self, workingDirectory, analysisDirectory, metaData)
        # Process the meta data.
        if metaData.has_key("minFrequencyOfInterestMultiplier"):
            self.minFrequencyOfInterestMultiplier=metaData["minFrequencyOfInterestMultiplier"]
        else:
            self.minFrequencyOfInterestMultiplier=0.7   
        if metaData.has_key("maxFrequencyOfInterestMultiplier"):
            self.maxFrequencyOfInterestMultiplier=metaData["maxFrequencyOfInterestMultiplier"]
        else:
            self.maxFrequencyOfInterestMultiplier=5            
        if metaData.has_key("recommendedFilterOrder"):
            self.filterOrder=metaData["recommendedFilterOrder"]
        else:
            self.filterOrder=5
        if metaData.has_key("peakThreshold"):
            self.peakThreshold=metaData["peakThreshold"]
        else:
            self.peakThreshold=0.05
        if metaData.has_key("plateSpecificMaxForceVariance"):
            self.plateSpecificMaxForceVariance=metaData["plateSpecificMaxForceVariance"]
        else:
            self.plateSpecificMaxForceVariance=5.0                
        if metaData.has_key("maxForceVariance"):
            self.maxForceVariance=metaData["maxForceVariance"]
        else:
            self.maxForceVariance=2.0
        if metaData.has_key("absurdForcePeak"):
            self.absurdForcePeak=metaData["absurdForcePeak"]
        else:
            self.absurdForcePeak=0.1
        if metaData.has_key("maxPeakRangeFraction"):
            self.maxPeakRangeFraction=metaData["maxPeakRangeFraction"]
        else:
            self.maxPeakRangeFraction=0.3                                 
        if metaData.has_key("phaseFrequencyRange"):
            self.phaseFrequencyRange=metaData["phaseFrequencyRange"]
        else:
            self.phaseFrequencyRange=0.1       
        if metaData.has_key("splitFraction"):
            self.splitFraction=metaData["splitFraction"]
        else:
            self.splitFraction=0.8
        # Standard calculations so our object is populated.
        self._calculateResolvents()
        self._calculateTotalForces()
        self._estimateStrideFrequency()
        self._selectPeaks()
        self._findTroughs()
        self._refinePeaks()
        self._plateSpecificPeaksAndTroughs()
        self._classifySignalType()
        self._orderPlates()
        self._findTotalDoubleIntegratedForce()
    
    def mangledKey(self,k):
        '''
        Overrides the base method and converts passed tuples to the 
        standard form for identifying force plates. 
        '''
        if type(k)==tuple:
            if len(k)>1:
                return(conv.ForcePlateIdentifierTuple(plate=k[0],direction=k[1]))
            else:
                return(k[0])
        return(k)
    
def getForcePlateWeightObject():
    '''
    Returns a sample force plate weight object
    '''
    workingDir = osjoin("..","sampleData","dogSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','forcePlateSampleWeightTrialDog.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(ForcePlateWeightData(workingDir,analysisDir,metaData["forcePlateData"]))

def getSampleForcePlateObject():
    '''
    Returns a sample force plate object from some dog data
    '''
    workingDir = osjoin("..","sampleData","dogSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','forcePlateSampleDataDog.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(ForcePlateData(workingDir,analysisDir,metaData["forcePlateData"]))

if __name__=='__main__':
    '''
    Runs some tests using the sample data
    '''
    print "Processing sample force plate data"
    sampleForcePlateData=getSampleForcePlateObject()
    sampleForcePlateData.makePlots()
    print "Processing sample force plate weight object"
    sampleForcePlateWeight=getForcePlateWeightObject()
    print "Weight of sample dog is ", sampleForcePlateWeight.weight, "N that is ", sampleForcePlateWeight.mass, "kg."
    sampleForcePlateWeight.makePlots()
