'''
Created on 17 May 2012

@author: aspence, aliedtke

A class which reads limb angle data and does phase based analyses.
Currently for nerve block gait project.
'''

import numpy as np
from phaser import Phaser
from scipy.stats import zscore
import csv, pylab
from matplotlib import font_manager
from os.path import join as osjoin
from os import listdir as listdir
from yaml import load as yamlload
from utils import angularCentralTendancy
import mpl_toolkits.mplot3d.axes3d as p3
from scipy.stats import nanstd
from scipy.stats import nanmean
from utils import FiltFiltFilter

from temporalSignal import TemporalSignal
from conventions import directory
#from baseData import BaseData
from directoryAware import DirectoryAware

class LimbAngleData(DirectoryAware):
    '''
    Limb angle data including methods to find estimates for the location of the 
    center of mass, leg phases, gaits, make various plots, detect footfalls and so on.
    '''
    defaultIDExt = "mcd"

    def _readData(self):
        '''
        Read a TSV file of limbAngleData.
        '''        
        # Find row right before data:
        reader = csv.reader(open(osjoin(self.workingDirectory, self.fileName), "r"),delimiter="\t")
        for row in reader:
            if reader.line_num == 4:
                print "Found the following keys:"
                print row
                labels = [x for x in map(unicode,row[0:])]
            elif reader.line_num > 3:
                break
                
        # Get the data:
        rdata = np.loadtxt(osjoin(self.workingDirectory, self.fileName),delimiter="\t",skiprows=4)
        print "Loaded raw data of size: "
        print rdata.shape
        self.totalSamples = rdata.shape[0]
        self.time=np.array(range(0,self.totalSamples))/float(self.sampleRate)
        
        # Now have rdata and labels, from any file format.
        # Put into a list of tuples, with index and labels in each tuple.
        labels=zip(range(0,len(labels)),labels)
        # Map over to our limbs notation for consistency with rough gait data:
        limbswap = {'leftfore':'FL','rightfore':'FR', 'lefthind':'BL','righthind':'BR'}
        # Loop over the index integer in i and label in k
        for i,k in labels:
            self.data[limbswap[k]]=TemporalSignal(rdata[:,i],self.time,0)
        
    def __setitem__(self, k, v):
        '''
        Virtual method to allow item assignment
        '''
        self.data[k]=v
        
    def mangleKey(self,k):
        '''
        Virtual method, when overridden should convert the key k to a
        sensible entry in the dictionary self.data. The idea is to make 
        the system robust by having it do something sensible regardless 
        of how the user request the data be given.
        Returns: A mangled version of the key k. 
        '''
        return(k)

    def has_key(self,k):
        '''
        Checks if the key k exists in the data object. The intention is to 
        allow users to query if a key exists interactively so the key is 
        mangled (projected from a list of acceptable key names on to the 
        space of actual key names in data) first. As such this should not 
        be used for the purposes of finding unique keys.
        Returns: If the key k corresponds to data in the object
        '''
        return(self.data.has_key(self.mangleKey(k)))

    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Returns: The data associated with k. 
        '''
        return(self.data[self.mangleKey(k)])

    def keys(self):
        '''
        Returns a list of unique keys for this data object. This is a list 
        of unique keys not a list of keys including mangling options. No mangling
        (projecting from a list of acceptable key names on to the space of actual
        key names in data) is performed here. 
        '''
        return(self.data.keys())
    
    def laPsec(self,x):
        filtFreq = 1.0
        filt = FiltFiltFilter(250,filtFreq,3) # Hardcoded, BAD!
        sig = filt.filter(x[0])
        return(sig)

    def _getPhases(self,phsr=None):
        '''
        Obtain the desired phases using phaser.
        
        Optional second argument is a tuple of a dictionnary of phasers with keys for limbs, to use to evaluate phases, and a string postfix to add to variable names in dictionnary.
        e.g. _getPhases((myphasersdict,"PreWalk"))
        
        NOTE ON WHAT PHASER EXPECTS:
        3D ARRAY:
          - TIME SERIES WITHIN MEASUREMENTS WITHIN TRIALS 
          ph = Phaser([np.array([zscore(np.array(d['FL']))])])
          y -- DxN or [ DxN_1, DxN_2, DxN_3, ... ] -- Measurements used for training
          - [1X TRIAL [1x meas [n time series] ]]
          - 
          - E.G. Make sure you pass either a numpy array that's DxN, where D is number of
          -  signals going into the estimate (e.g. gyro1, accel2, etc.) and N is the
          -  number of time points, OR
          -  a LIST containing some number of repeated sets of these, e.g. [ DxN_1, DxN_2 ... ]
          -  
          - To look for phase dependent change (which we might expect with nerve block):
          -   Idea  to plot leg angles versus global phase--with error enhance of phaser
          -  -compute global phase from zscoreed all four leg angles
          -    - one for each condition (blocked and no blocked)
          -    - use poinscare roughly equal to midstance of hind left. this is where angle changing most rapidly.
          -  - plot leg angles against their phases.
        '''
        print "Evaluating phases using phaser"
        pphase = dict()
        for limb in self.limbs:
            pphase[limb] = np.array([zscore(np.array(self[limb]))]) # np.array([zscore(np.array(d['FL']))])
            # Build Phaser
            if phsr is None:
                postfix = ""
                self[limb+"Phaser"] = Phaser(pphase[limb]) # By default, train on selfsame limb data, all of it
            else:
                postfix=phsr[1]
                self[limb+"Phaser"+postfix] = phsr[0][limb]
            # Get Phases
            self[limb+"Phase"+postfix] = self[limb+"Phaser"+postfix].phaserEval(pphase[limb])[0,:]%(2.0*np.pi)
        setattr(self,"relativePhaseLabels"+postfix,[l1+l2+"RelativePhase"+postfix for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
        for l1,l2 in zip(self.limbs[:-1],self.limbs[1:]):
            self[l1+l2+"RelativePhase"+postfix] = (self[l1+"Phase"+postfix]-self[l2+"Phase"+postfix])%(2.0*np.pi)
        for l1,l2 in zip(['FR','BR','BL'],['FL','FL','FL']):
            self[l2+l1+"PhaseDiff"+postfix]=(self[l2+"Phase"+postfix]-self[l1+"Phase"+postfix])%(2.0*np.pi)

        # Construct a global kinematic phase?
        #self.phaser = Phaser( np.vstack([self[k] for k in self.limbs]), psecfunc=self.laPsec )
        #self["globalPhase"] = self.phaser.phaserEval(np.vstack([self[k] for k in self.limbs]))[0,:]%(2.0*np.pi)
        
    def _strideCut(self,cutpostfix="",storepostfix=""):
        # Find strides based on fore left limb phase
        # More things to compute: 
        # - Use angularCentral on striderelatives to compute a "middle" of the cloud
        # - use g3D to compute distances of all poitns to centroid
        # - Compare distribution of distances for various conditions. 
        # -  - compute distance with (point - centroid) dot g3D dot (point - centroid)
        self["relativeStridePhases"+storepostfix]=list()
        self.strideIdx = [idx for e1,e2,idx in zip((self["FLPhase"+cutpostfix][:-1]+np.pi)%(2.0*np.pi)-np.pi,(self["FLPhase"+cutpostfix][1:]+np.pi)%(2.0*np.pi)-np.pi,range(len(self["FLPhase"+cutpostfix][:-1]))) if e1<0 and e2>0]
        self.firstStrideIdx = self.strideIdx[self.firstStride]
        self.lastStrideIdx = self.strideIdx[self.lastStride]
        for idx0,idx1 in zip(self.strideIdx[self.firstStride:self.lastStride-1],self.strideIdx[self.firstStride+1:self.lastStride]): # HARDCODED to ignore the first few
            self["relativeStridePhases"+storepostfix].append([
                                                   angularCentralTendancy(self[l1+l2+"RelativePhase"+storepostfix][idx0:idx1],ctf=nanmean)
                                                 for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
        self["relativeStridePhases"+storepostfix] = np.array(self["relativeStridePhases"+storepostfix])
        setattr(self,"relativeStridePhaseLabels"+storepostfix,[l1+l2+"RelativeStridePhase"+storepostfix for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
        
    def _computeCentroids(self,postfix=""):        
        # Compute centroid of whole waveform relative phases:
        self["relativePhaseCentroid"+postfix] = [angularCentralTendancy(self[l],ctf=nanmean) for l in getattr(self,"relativePhaseLabels"+postfix)]
        # Compute centroid of stride relative phases:
        self["relativeStridePhaseCentroid"+postfix] = [angularCentralTendancy(self["relativeStridePhases"+postfix][:,k],ctf=nanmean) for k in range(len(getattr(self,"relativePhaseLabels"+postfix)))]
        setattr(self,"relativeStridePhaseCentroidLabels"+postfix,[l1+l2+"RelativeStridePhaseCentroid"+postfix for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
            
    def plotPhaseComputation(self):
        pylab.figure(figsize=(13,7.5))
        for j,limb in enumerate(self.limbs):
            ax=pylab.subplot(1,4,j+1)
            wid = 5
            labs = ['leg angle','phase']
            locs = np.array(range(len(labs)))*5
            mid = np.mean([self.time[0],self.time[-1]])
            pylab.plot(self.time-mid,zscore(self[limb])+locs[0])
            pylab.plot(self.time-mid,self[limb+'Phase']+locs[1])
            ax.set_yticks(locs)
            ax.set_yticklabels(labs,rotation=90)
            ax.set_xlim([-wid/2,+wid/2])
            pylab.grid()
            if j==0:
                ax.set_ylabel('Signal')
            ax.set_xlabel('Time (sec)')
            ax.set_title(limb)
            ax.set_ylim((-5,5*(len(labs)+1)))
        #ax.annotate( " ".join([self.parent.parent.iD,self.parent["trialType"], \
        #    self.parent["terrainType"],self.parent.iD]),xy=(0.5,1), xytext=(0,-10),\
        #    xycoords='figure fraction',textcoords='offset points',va='top',ha='center')
        #pylab.savefig(osjoin(self.analysisDirectory, "_".join(["phaseComputation", \
        #    self.parent.parent.iD,self.parent["trialType"],self.parent["terrainType"], \
        #    self.parent.iD,".png"])))
        pylab.close('all')
        #pylab.show()
    
    def plotRelativePhases(self):
        sigs = list()
        # this is first limb to second from last zipped with second limb through last:
        #In [19]: zip(imud.limbs[:-1],imud.limbs[1:])
        #Out[19]: [('FL', 'BR'), ('BR', 'BL'), ('BL', 'FR')]
        labs = list()
        for l1,l2 in zip(self.limbs[:-1],self.limbs[1:]):
            sigs.append(self[l1+l2+"RelativePhase"])
            labs.append(l1+l2+"RelPh")
        rph = np.array(sigs).T
        pylab.figure(figsize=(13,7.5))
        #ax1 = pylab.subplot2grid((4,1), (0, 0), rowspan=3)
        pylab.plot(self.time-self.time[0],rph)
        ax1 = pylab.gca()
        pylab.legend(labs)
        pylab.hlines([np.pi/2, np.pi,3*np.pi/2,2*np.pi],0,self.time[-1]-self.time[0])
        ax1.set_ylabel('Relative leg phase (rad)')
        ax1.set_xticks([])
        #ax1.set_title(" ".join([self.parent.parent.iD,self.parent["trialType"], \
        #    self.parent["terrainType"],self.parent.iD]))
        ax1.set_ylim([0, np.pi*2])
        ax1.set_xlim((0,self.time[-1]-self.time[0]))
        
        #pylab.show()
        #pylab.savefig(osjoin(self.analysisDirectory, "_".join(["relativePhases", \
        #    self.parent.parent.iD,self.parent["trialType"],self.parent["terrainType"], \
        #    self.parent.iD,".png"])))

    def plot3DPhaseScatterPlot(self):
        x=self["relativeStridePhases"].transpose()[0]
        y=self["relativeStridePhases"].transpose()[1]
        z=self["relativeStridePhases"].transpose()[2]
        fig=pylab.figure()
        ax = p3.Axes3D(fig)
        ax.scatter3D(np.ravel(x),np.ravel(y),np.ravel(z),c='r')
        ax.set_xlabel('Phi1')
        ax.set_ylabel('Phi2')
        ax.set_zlabel('Phi3')
        ax.set_xlim3d([0,2.0*np.pi])
        ax.set_ylim3d([0,2.0*np.pi])
        ax.set_zlim3d([0,2.0*np.pi])
        fig.add_axes(ax)
        #fig.legend([(l1+l2+"RelativePhase") for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
                     
                    
    def plotData(self):
        pylab.figure()
        pylab.clf()
        [pylab.plot(self[k]) for k in self.keys()]
        pylab.title("limbAngleData for %s" % self.fileName )
        pylab.xlabel("sample")
        pylab.ylabel("angle")
        pylab.legend(self.keys(),loc=1,prop=font_manager.FontProperties(size=6))
        pylab.show()
             
    def writeData(self):
        '''
        Write data at limb angle level
        '''
        print "TODO"
                  
    def __init__(self, workingDirectory, analysisDirectory, metaData ):
        '''
        Constructor - Runs through the procedure of calculating computationall cheap parameters
        such as angles and lengths
        '''
        DirectoryAware.__init__(self, workingDirectory,analysisDirectory,metaData)
        self.fileName = metaData["fileName"]
        self.sampleRate = metaData["sampleRate"]
        self.limbs=["FL","BR","BL","FR"]
        self.gait = metaData["gait"]
        self.data = dict()
        if metaData.has_key("firstStride"):
            self.firstStride = metaData["firstStride"]
            if metaData.has_key("numStrides"):
                self.numStrides = metaData["numStrides"]
                self.lastStride = self.firstStride + self.numStrides
            else:
                self.lastStride = metaData["lastStride"]
        else:
            self.firstStride = 2
            self.lastStride = -2
        self._readData()
        self._getPhases()
        self._strideCut()
        self._computeCentroids()
        #self.plotData()
        #self.plotPhaseComputation()
        #self.plotRelativePhases()
        #self.plot3DPhaseScatterPlot()
        
if __name__=='__main__':
    '''
    Runs some tests using the sample data
    '''
    print "TODO: Run tests here."