'''
Created on 18 Jan 2011

@author: Simon Wilshin, swilshin@rvc.ac.uk

A class which reads motion capture data from files or directories and then 
provides easy, dictionary like interactions with the data. Can read multiple 
different file types and will generally make smart choices about what a file 
type means
'''

import numpy as np
import csv
from os.path import join as osjoin
from os import listdir as listdir
from yaml import load as yamlload
from xlrd import open_workbook
from scipy.io import loadmat
from utils import ensure_dir

from temporalSignal import TemporalSignal
from conventions import directory
from baseData import BaseData

import pylab as pl

class MotionCaptureData(BaseData):
    '''
    Motion capture data including methods to find estimates for the location of the 
    center of mass, leg phases, gaits, make various plots, detect footfalls and so on.
    '''
    defaultIDExt = "mcd"
    def _readXLSDir(self):
        '''
        Read a directory of XLS files stored in accordance with the format of 
        Alexis' fish data.
        '''
        rdata=list()
        labels=list()
        for f in listdir(osjoin(self.workingDirectory,self.fileName)):
            if f[-3:]=="xls":
                nd=open_workbook(osjoin(self.workingDirectory,self.fileName,f))
                nd=nd.sheet_by_index(0)
                labels.append(f[:-4])
                nrdata=list()
                for rownum in range(1,nd.nrows):
                    rd=nd.row_values(rownum)
                    nrdata.append([float(rd[2]),float(rd[3]),0])
                rdata.append(np.array(nrdata).transpose())
        rdata=np.array(rdata)
        return(rdata,labels)
    
    def _readQTMMat(self):
        '''
        Read a MATLAB mat file as exported from QTM.
        '''
        rdata=loadmat(osjoin(self.workingDirectory, self.fileName),struct_as_record=False)
        if self.sampleRate!=None:
            print "WARNING: Sample Frequency found in motion capture data file. OVERWRITING YAML ENTRY."
            print "  YAML SampleRate: %d MoCap File SampleRate: %f" %  (self.sampleRate,float(rdata[self.qualisysNames][0,0].FrameRate[0,0]))                 
        self.sampleRate=float(rdata[self.qualisysNames][0,0].FrameRate[0,0])                
        # This is an array of u strings:
        labels=[ l[0] for l in rdata[self.qualisysNames][0,0].Trajectories[0,0].Labeled[0,0].Labels[0][:]]
        # This is 3D ndarray of [markers xyzresid samples] e.g. shape [11 4 3000]
        # Strip off residuals (2nd dim :3 below). This is confusing and we don't need it.
        rdata=rdata[self.qualisysNames][0,0].Trajectories[0,0].Labeled[0,0].Data[:,:3,:]
        return(rdata,labels)

    def _readQTMTSV(self):
        '''
        Read a TSV file as exported from QTM.
        '''        
        # Find row right before data:
        reader = csv.reader(open(osjoin(self.workingDirectory, self.fileName), "r"),delimiter="\t")
        br=False
        changed=False
        for row in reader:
            # Each row returned as a list, each element is the string in that tab-delimited entry:
            if row[0]=="MARKER_NAMES":
                nameslinenum = reader.line_num
                # Could store marker types in a dictionary somewhere
                #print {row[0]:row[1:len(row)]}
                # Need to create an array of unicode strings:
                # This creates a list of unicode strings, which works with zip below.
                labels = [x for x in map(unicode,row[1:])]
                br=True
                changed=True
            elif row[0]=="Frame":
                nameslinenum = reader.line_num
                changed=True
                br=True
            elif row[0]=="FREQUENCY":
                if self.sampleRate!=None:
                    print "WARNING: Sample Frequency found in motion capture data file. OVERWRITING YAML ENTRY."
                    print "  YAML SampleRate: %d MoCap File SampleRate: %s" %  (self.sampleRate,row[1]) 
                self.sampleRate = float(row[1])
                # br and changed should be false and stay that way
                # as this item will be above either MARKER_NAMES or Frame:
            else:
                changed=False
            if br and not changed:
                break
        # Get the data:
        rdata = np.loadtxt(osjoin(self.workingDirectory, self.fileName),delimiter="\t",skiprows=nameslinenum)
        if np.shape(rdata)[1]%3!=0:
            rdata=rdata[:,2:]
        # This is now [samples 3xMARKERSxyz]. To work with below need to reshape, to 3D.
        # np reshape appears to fill row - wise, not column wise. This will be [mrkrs xyz samples]
        rdata = rdata.transpose().reshape(np.shape(rdata)[1]/3,3,np.shape(rdata)[0])
        return(rdata,labels)
        
    def _readData(self):
        '''
        Reads data exported from Qualisys Track Manager in two different formats. The ".mat" format 
        (MATLAB) and the tab seperated value format. Gets the track labels to use as dictionary keys. 
        Use sensible names for your markers and things should 'just work'.
        
        This should always ignore Qualisys residuals since nothing is ever done with them and they 
        aren't exactly trustworthy.
        '''
        if self.fileType==directory:
            rdata,labels = self._readXLSDir()
        else:
            if self.fileName[-3:] == "mat":
                rdata,labels = self._readQTMMat()
            elif self.fileName[-3:] == "tsv":
                rdata,labels = self._readQTMTSV()
        if self.sampleRate==None:
            print "WARNING: SAMPLE RATE NOT SET!"
        # Calculate a vector of times, if you don't know how long the 
        # samples are guess based on the first estimate
        if self.nanTrim and not self.timeCuts: # Cant do both, default to doing as instructed in cuts
            # Remove Nans from start and end
            g = np.nonzero(np.sum(np.sum(~np.isnan(rdata),0),0))[0]
            rdata = rdata[:,:,g[0]:g[-1]]
        if self.timeCuts:
            minIdx = int(self.timeCuts[0]*self.sampleRate)
            if minIdx < 0:
                minIdx = 0
            maxIdx = int(self.timeCuts[1]*self.sampleRate)+1
            if maxIdx >= rdata.shape[2]:
                maxIdx = rdata.shape[2]-1
            rdata = rdata[:,:,minIdx:maxIdx]
        if not self.totalSamples:
            self.totalSamples=rdata[0].shape[1]
        self.time=np.array(range(0,self.totalSamples))/float(self.sampleRate)
        
        # Store the labels as they come from raw data, file, so can be used as keys to get raw data.
        self.labels=labels
        
        # Now have rdata and labels, from any file format.
        # Put into a list of tuples, with index and labels in each tuple.
        labels=zip(range(0,len(labels)),labels)
        # Loop over the index integer in i and label in k
        converted=False
        for i,k in labels:
            if self.zeroToNan:
                if not converted:
                    print "Converting 0.000 to Nan as specified."
                    converted=True
                rdata[i][rdata[i]==0.000]=np.NaN
            if self.flipX:
                if self.metaData["verbosity"]>0:
                    print "Flipping sign of X coordinate as specified"
                rdata[i][0] = -rdata[i][0]
            # Note this will use __setitem__ virtual method of BaseData.
            self[k]=TemporalSignal(rdata[i],self.time,0)
        
    def _getDirections(self):
        '''
        Find the directions which correspond to forward, vertical and left to right 
        for our animal. At present this assumes x is fore-aft, y is medilateral, 
        z is vertical. Should have a better way of doing this.
        BUG0006 - Make this use body co-ordinates
        '''
        print "Calculating directions of motion..."
        self.directionLabels=[u"foreAft",u"medilateral",u"vertical"]
        self[u"foreAft"]=np.array([1,0,0])
        self[u"medilateral"]=np.array([0,1,0])
        self[u"vertical"]=np.array([0,0,1])
       
    def __init__(self, workingDirectory, analysisDirectory, metaData, parent):
        '''
        Constructor - Runs through the procedure of calculating computationall cheap parameters
        such as angles and lengths
        '''
        # Note that BaseData handles sampleRate, but that readData
        # will override sampleRate given in yaml if it conflicts with
        # one found in the data file.
        BaseData.__init__(self, workingDirectory, analysisDirectory, metaData, parent)
        if self.metaData.has_key("trialType"):
            self.trialType=self.metaData["trialType"]
        else:
            self.trialType="overground"
        if self.metaData.has_key("zeroToNan"):
            self.zeroToNan=self.metaData["zeroToNan"]
        else:
            self.zeroToNan=False   
        if self.metaData.has_key("qualisysNames"):
            self.qualisysNames=self.metaData["qualisysNames"]
        else:
            self.qualisysNames=self.fileName[:-4]
        if self.metaData.has_key("nanTrim"):
            self.nanTrim=self.metaData["nanTrim"]
        else:
            self.nanTrim=False
        if self.metaData.has_key("mocapFlipX"):
            self.flipX=True
        else:
            self.flipX=False
        self._readData()

    def zerosToNans(self):
        if self.metaData["verbosity"]>0:
            print "Mocap %s converting zero to nan in raw data" % self.iD
        for k in self.labels:
            self.data[k][self.data[k]==0.000] = np.nan

    def nansByMarker(self):
        if self.metaData["verbosity"]>1:
            print "Mocap %s counting nans in data" % self.iD
        nancnt=dict()
        for k in self.labels:
            nancnt[k]=int(np.sum(np.isnan(self.data[k])))
        return nancnt
    
    def plotMarkersWithZeros(self,tofile=True,targetdir=None):
        if self.hasKeyAndTrue("zeroToNan"):
            print "Warning: call to plotMarkersWithZeros pointless as zeroToNan flag set in metaData."
            return
        else:
            # Problem: self[k] will call derived method that may interpolate data. Avoid by accessing
            # data directly with stored raw data labels. Derived classes override __getitem__ and 
            # grab items from data with postpended _gpint etc.
            if targetdir is None:
                targetdir=osjoin(self.analysisDirectory,"zerochecking")
            for k in self.labels:
                if np.any( self.data[k]==0.000 ):
                    pl.figure()
                    pl.plot(self.data[k].T)
                    pl.legend(('x','y','z'))    			
                    if tofile:
                        ensure_dir(targetdir)
                        targetfile = osjoin(targetdir, ".".join(["_".join([self.fileName[:-4],k,"haszeros"]),"png"]))
                        print "  Writing image %s" % targetfile
                        pl.savefig(targetfile)
                        #pl.close()               

    def plotMarkersWithZerosOrNans(self,tofile=True,targetdir=None,suffix="haszerosornans"):
        if targetdir is None:
            targetdir=osjoin(self.analysisDirectory,"zerosornans")
        for k in self.labels:
            if np.any( self.data[k]==0.000 ) or np.any( np.isnan(self.data[k]) ):
                pl.figure()
                pl.plot(self.data[k].T)
                pl.legend(('x','y','z'))    			
                if tofile:
                    ensure_dir(targetdir)
                    targetfile = osjoin(targetdir, ".".join(["_".join([self.fileName[:-4],k,suffix]),"png"]))
                    print "  Writing image %s" % targetfile
                    pl.savefig(targetfile)
    
    def plotRawMocap(self,tofile=True,targetdir=None,withnans=False,marker=None):
        if targetdir is None:
            targetdir=self.analysisDirectory
        if marker is None:
            markers=self.labels
        else:
            if type(marker)==str:
                markers=[marker]
            else:
                markers=marker
        for k in markers:
            sigs = self.data[k]
            if withnans:
                if not np.any( np.isnan(self.data[k]) ):
                    break
            pl.figure()
            pl.plot(sigs.T)
            pl.title(self.fileName[:-4])
            if tofile:
                ensure_dir(targetdir)
                targetfile = osjoin(targetdir, ".".join(["_".join([self.fileName[:-4],k,"arawdata"]),"png"]))
                print "  Writing image %s" % targetfile
                pl.savefig(targetfile)
                pl.close()        
        
    def plotFill(self,tofile=True,targetdir=None,suffix="nanfill",marker=None):
        if targetdir is None:
            targetdir=self.analysisDirectory
        if marker is None:
            markers=self.labels
        else:
            if type(marker)==str:
                markers=[marker]
            else:
                markers=marker
        for k in markers:
            sigs = self.data[k]
            if np.any( np.isnan(self.data[k]) ):
                if self.metaData["verbosity"]>0:
                    print "  Mocap object iD %s label %s has %d nans" % (self.iD,k,np.sum(np.isnan(self.data[k]))) 
                pl.figure()
                pl.plot(sigs.time,sigs.T)
                pl.plot(sigs.time,self[k].T)
                pl.plot(sigs.time[np.where(np.isnan(sigs[0]))[0]],self[k][:,np.where(np.isnan(sigs[0]))[0]].T,'o')
                pl.title(self.fileName[:-4])
                if tofile:
                    ensure_dir(targetdir)
                    targetfile = osjoin(targetdir, ".".join(["_".join([self.fileName[:-4],k,suffix]),"png"]))
                    print "  Writing image %s" % targetfile
                    pl.savefig(targetfile)
                    pl.close()
                else:
                    pl.show()

    def plotMarkerFill(self,k,tofile=True,targetdir=None,suffix="nanfill"):
        if targetdir is None:
            targetdir=self.analysisDirectory
        sigs = self.data[k]
        if np.any( np.isnan(self.data[k]) ):
            if self.metaData["verbosity"]>0:
                print "  Mocap object iD %s label %s has %d nans" % (self.iD,k,np.sum(np.isnan(self.data[k]))) 
            pl.figure()
            pl.plot(sigs.T)
            pl.plot(self[k].T)
            pl.plot(np.where(np.isnan(sigs[0]))[0],self[k][:,np.where(np.isnan(sigs[0]))[0]].T,'o')
            if tofile:
                ensure_dir(targetdir)
                targetfile = osjoin(targetdir, ".".join(["_".join([self.fileName[:-4],k,suffix]),"png"]))
                print "  Writing image %s" % targetfile
                pl.savefig(targetfile)
                pl.close()
        
def getSampleMotionCaptureObject():
    '''
    Returns a sample motion capture data object based on some 
    sample data from dogs
    '''
    workingDir = osjoin("..","sampleData","dogSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','motionCaptureSampleDataDog.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(MotionCaptureData(workingDir,analysisDir,metaData,None))

def getSampleTSVMotionCaptureObject():
    '''
    This function returns an example of a motion capture dataModule
    object, based on a TSV data file exported from qualisys.
    '''
    workingDir = osjoin("..","sampleData","dogSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','motionCaptureSampleDataDogTSV.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(MotionCaptureData(workingDir,analysisDir,metaData,None))

def getSampleHorseTSVMotionCaptureObject():
    '''
    This function returns an example of a motion capture dataModule
    object, based on a TSV data file exported from qualisys from a horse.
    '''
    workingDir = osjoin("..","sampleData","horseSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','motionCaptureSampleDataHorseTSV.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()   
    return(MotionCaptureData(workingDir,analysisDir,metaData,None))

def getSampleHorseGaitTransition(idx):
    '''
    This function returns an example of a motion capture dataModule
    object, based on a TSV data file exported from qualisys. It uses
    data from the jonny_20101112_nb0001_an_sub.tsv file which should 
    contain a gait transition.
    '''
    workingDir = osjoin("..","sampleData","horseSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = [
                     osjoin("..",'sampleYAML','dataContainers','gaitTransitionHorseSampleWalkToTrot.yaml'),
                     osjoin("..",'sampleYAML','dataContainers','gaitTransitionHorseSampleTrotToWalk.yaml'),
                   ]
    stream = file(metaDataFile[idx], 'r')
    metaData=yamlload(stream)
    stream.close()  
    return(MotionCaptureData(workingDir,analysisDir,metaData,None))
 
def getSampleDogGaitTransition():
    '''This function returns an example of a motion capture dataModule
    object, based on a TSV data file exported from qualisys. It uses
    data from the jonny_20101112_nb0001_an_sub.tsv file which should 
    contain a gait transition.'''
    workingDir = osjoin("..","sampleData","dogSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','gaitTransitionSampleDataDog.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()    
    return(MotionCaptureData(workingDir,analysisDir,metaData,None))

def getSampleFishMotionCaptureObject():
    '''This function returns an example of a motion capture dataModule
    object, based on the directory structure of Alexis' data from fish. 
    '''
    workingDir = osjoin("..","sampleData","fishSampleData")    
    analysisDir = osjoin("..","sampleAnalysis","dataContainers")
    metaDataFile = osjoin("..",'sampleYAML','dataContainers','motionCaptureSampleDataFish.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(MotionCaptureData(workingDir,analysisDir,metaData,None))  

if __name__=='__main__':
    '''
    Runs some tests using the sample data
    '''
    print "Processing sample motion capture data (dog, mat)"
    a=sampleQualisysData=getSampleMotionCaptureObject()
    print "Processing sample motion capture data (dog, tsv)"
    b=sampleQualisysDataTSV=getSampleTSVMotionCaptureObject()
    print "Processing sample motion capture data (horse, mat)"
    c=sampleHorseTSVData = getSampleHorseTSVMotionCaptureObject()
    print "Processing sample motion capture data with gait transition (horse, first sample)"
    d=sampleTransitionData = getSampleHorseGaitTransition(0)
    print "Processing sample motion capture data with gait transition (horse, second sample)"    
    e=sampleTransitionData = getSampleHorseGaitTransition(1)
    print "Processing sample motion capture data with gait transition (dog)"    
    f=sampleTransitionData = getSampleDogGaitTransition()
    print "Processing sample motion capture data (fish)"    
    g=sampleFishData = getSampleFishMotionCaptureObject()
