'''
Created on 22 May 2012

@author: aspence

A class which reads tarsus tracking data from a .mat file and computes
leg phases, relative leg phases, and some gait distances.

TODO: For Ryan Harris:
Work backwards from what we want to compute.
- Intact spiders:
    - Leg phases
      - How to handle nan?
    - Relative leg phases
    - Stride averaged relative leg phases
      - Which leg to cut on?
    - Use these to compute distances to gaits.
    - Supply a dictionnary of gaits. Code automagically computes distance to all of these, and exports.
      - Which object should this really be in?

Let's work through the legs, and what we will get.

Using Paul's suggestion as it makes alternating tetrapods easy.
4 3 2 1
------>
5 6 7 8
alt pods are then 1-3-5-7, 2-4-6-8

Thus the phase differences for legs will be, using the Simon relative leg phase convention
alttet = [ 1-2 = pi, 2-3 = pi, 3-4 = pi, 4-5 = pi, 5-6 = pi, 6-7 = pi, 7-8 = pi ]
Intact alt tet = [pi pi pi pi pi pi pi]
- Ablated spiders: we need to add one more relative leg phase, and then can distinguish tripods and alttest (which become vaults):
We need to go down to a five torus, as we have six legs. Just add the 3-5 rel leg phase column, and we can do:
Note: Ryan abaltes legs 4 and 8 in this notation (columns 1 and 5 in matlab)
x 3 2 1
------>
5 6 7 x
Limbs left in this notation = [1,2,3,5,6,7]
And then alternating tetrapod (phsae 0 1357,phase pi 2468) becomes the vault (1357,26), and if we use relative leg phases of
1-2 2-3 3-5 5-6 6-7
pi   pi  0   pi  pi
But alternating tripod is in phase groups (3 6 1) and (5 2 7)
pi   pi  pi  pi  pi

TODO: Make this object handle two trial types from metadata.
TODO:   - Only compute phases for legs you have
TODO:   - Compute the correct set of leg relative phases
TODO:   - 
'''

import numpy as np
from phaser import Phaser
from scipy.stats import zscore
import csv, pylab
from matplotlib import font_manager
from os.path import join as osjoin
from os import listdir as listdir
from os.path import join as osjoin
from yaml import load as yamlload
from utils import angularCentralTendancy
import mpl_toolkits.mplot3d.axes3d as p3
from scipy.stats import nanstd
from scipy.stats import nanmean
from scipy.io import loadmat
from scipy.interpolate import splrep, splev
from utils import FiltFiltFilter

from temporalSignal import TemporalSignal
from conventions import directory
#from baseData import BaseData
from directoryAware import DirectoryAware

class SpiderTarsiTrackData(DirectoryAware):
    '''
    Load matlab file from miscTracker used to track tarsi of a spider..
    '''
    defaultIDExt = "mcd"

    def _readData(self):
        '''
        Read a .mat file of spider tarsus tracking data.
        '''        
        mdata=loadmat(osjoin(self.workingDirectory, self.fileName),struct_as_record=False)
        self.res = mdata['res'][0,0] # Matlab structs actually are arrays fo struct - get the 0th element.
        # res is now a python object with data items ini, pthX, pthY. can plot with self.plotData
        # now, do we convert data? put into dictionnary with keys?
        # Yes. 1) I'm used to this paradigm. 2) It is self-documenting.
        
        '''
        Columns of pthX, pthY, as digised by Ryan are:
        5 6 7 8    or in python 4 5 6 7
        ------>       
        4 3 2 1                 3 2 1 0
        
        Using Paul's suggestion as it makes alternating tetrapods easy.
        4 3 2 1
        ------>
        5 6 7 8
        alt pods are then 1-3-5-7, 2-4-6-8

        Thus the phase differences for legs will be, using the Simon relative leg phase convention
        alttet = [ 1-2 = pi, 2-3 = pi, 3-4 = pi, 4-5 = pi, 5-6 = pi, 6-7 = pi, 7-8 = pi ]
        
        Note that for ablated files, columns 1 and 5, for legs one and five, should be ignored. These
        columns should be zeros or nans.
        
        UHOH: Looing at the data: Ryan has ablated columns 1 and 5 in his convention, which is actually
        legs FRONT RIGHT and BACK LEFT. So in Paul's world, this is legs 8 and 4. FIXED.
        
        '''
        # Map Ryan's data into Pauls' convention:
        if self.trialType == "intact":
            self.legmap = [8,7,6,5,4,3,2,1]
        elif self.trialType == "ablated":
            # Limbs here is 1,2,3,5,6,7 in Paul notation: which columns in matlab?
            self.legmap = [8,7,6,4,3,2]
        # k goes 0,1,...7. so in intact limbs 1,2,3, get columns 7,6,5,4,3,2,1,0
        # ablated goes limbs 1,2,3,5,6,7 should get columns 7,6,5,3,2,1 (python zero based). this is correct.
        for k in range(len(self.limbs)):
            self.data[self.limbs[k]]=self.res.pthX[self.trimStart:self.trimEnd,self.legmap[k]-1]
        # Note the below works (has trimmed number of samples) because __getitem__ will return one of
        # the newly set data dict items from the above.
        self.time = np.array(range(len(self[self.limbs[0]]))) / float(self.sampleRate)
        #self.time = np.array(range(0,self.totalSamples))/float(self.sampleRate)
        
    def __setitem__(self, k, v):
        '''
        Virtual method to allow item assignment
        '''
        self.data[k]=v
        
    def mangleKey(self,k):
        '''
        Virtual method, when overridden should convert the key k to a
        sensible entry in the dictionary self.data. The idea is to make 
        the system robust by having it do something sensible regardless 
        of how the user request the data be given.
        Returns: A mangled version of the key k. 
        '''
        return(k)

    def has_key(self,k):
        '''
        Checks if the key k exists in the data object. The intention is to 
        allow users to query if a key exists interactively so the key is 
        mangled (projected from a list of acceptable key names on to the 
        space of actual key names in data) first. As such this should not 
        be used for the purposes of finding unique keys.
        Returns: If the key k corresponds to data in the object
        '''
        return(self.data.has_key(self.mangleKey(k)))

    def __getitem__(self,k):
        '''
        Returns an element of the dictionary self.data based on the key k 
        which is mangled (projected from a list of acceptable key names on 
        to the space of actual key names in data) first so the user has a 
        variety of sensible ways to get at the data. Will throw an exception 
        if k does not correspond to anything after mangling.
        Returns: The data associated with k. 
        '''
        return(self.data[self.mangleKey(k)])

    def keys(self):
        '''
        Returns a list of unique keys for this data object. This is a list 
        of unique keys not a list of keys including mangling options. No mangling
        (projecting from a list of acceptable key names on to the space of actual
        key names in data) is performed here. 
        '''
        return(self.data.keys())
    
    def laPsec(self,x):
        filtFreq = 1.0
        filt = FiltFiltFilter(250,filtFreq,3) # Hardcoded, BAD!
        sig = filt.filter(x[0])
        return(sig)

    def _getPhases(self):
        '''
        Obtain the desired phases using phaser.
        
        NOTE ON WHAT PHASER EXPECTS:
        3D ARRAY:
          - TIME SERIES WITHIN MEASUREMENTS WITHIN TRIALS 
          ph = Phaser([np.array([zscore(np.array(d['FL']))])])
          y -- DxN or [ DxN_1, DxN_2, DxN_3, ... ] -- Measurements used for training
          - [1X TRIAL [1x meas [n time series] ]]
          - 
          - E.G. Make sure you pass either a numpy array that's DxN, where D is number of
          -  signals going into the estimate (e.g. gyro1, accel2, etc.) and N is the
          -  number of time points, OR
          -  a LIST containing some number of repeated sets of these, e.g. [ DxN_1, DxN_2 ... ]
          -  
          - To look for phase dependent change (which we might expect with nerve block):
          -   Idea  to plot leg angles versus global phase--with error enhance of phaser
          -  -compute global phase from zscoreed all four leg angles
          -    - one for each condition (blocked and no blocked)
          -    - use poinscare roughly equal to midstance of hind left. this is where angle changing most rapidly.
          -  - plot leg angles against their phases.
        '''
        print "Evaluating phases using phaser"
        pphase = dict()
        trainpphase=dict()
        for limb in self.limbs:
            if self.debug:
                pylab.figure()
                pylab.plot(self[limb])
            # Zscore the 1D limb data for this trial
            limbzsc = np.array([zscore(np.array(self[limb]))])
            #print limbzsc.shape # (1,218)
            # Get Phaser
            if False:
                pphase[limb] = limbzsc # np.array([zscore(np.array(d['FL']))])
                filt = FiltFiltFilter(float(self.sampleRate),50.0,3)
                pphase[limb] = filt(pphase[limb])
                #print pphase[limb]
                self[limb+"PhaseSig"]=pphase[limb][0]
                pylab.plot(pphase[limb][0])
            else:
                if self.traindata is None:
                    #print limbzsc
                    #print limbzsc.shape
                    x=range(limbzsc[0].shape[0])
                    # Played around with smoothing coeffs: 500 looks good for heavy lowpass
                    # Ah. The smoothing depends on sample size. 
                    tck=splrep(x,limbzsc[0],w=np.ones((limbzsc[0].shape[0],1)).fill(0.1),s=(limbzsc[0].shape[0]/self.smoothfactor))
                    pphase[limb]=np.array([splev(x,tck)])
                    self.phaser = Phaser(pphase[limb])
                    #pylab.plot(pphase[limb][0])
                    #pylab.show()
                    #print pphase[limb]
                    self[limb+"PhaseSig"]=pphase[limb][0]
                    self[limb+"Phase"] = self.phaser.phaserEval(pphase[limb])[0,:]%(2.0*np.pi)
                else:
                    # We potentially have multipel training sequecnes to use. Train the phsaer using the set of training sequences set in self.traindata by teh experiment object;
                    trainpphase[limb]=list()
                    for seq in self.traindata[limb]:
                        #print seq.shape # (460,)
                        seqzsc=np.array([zscore(np.array(seq))])
                        #print seqzsc.shape #(1,460) - this is 2D array with 1 element in first dimension, 460 in second. zscore?
                        x=range(seqzsc[0].shape[0]) #[0,1,2...]
                        tck=splrep(x,seqzsc[0],w=np.ones((seqzsc[0].shape[0],1)).fill(0.1),s=(seqzsc[0].shape[0]/self.smoothfactor))
                        trainpphase[limb].append(np.array([splev(x,tck)]))
                    self.phaser = Phaser(trainpphase[limb])
                    # Prepare teh phase signal for this individual limb:
                    x=range(limbzsc[0].shape[0])
                    # Played around with smoothing coeffs: 500 looks good for heavy lowpass
                    # Ah. The smoothing depends on sample size. 
                    tck=splrep(x,limbzsc[0],w=np.ones((limbzsc[0].shape[0],1)).fill(0.1),s=(limbzsc[0].shape[0]/self.smoothfactor))
                    pphase[limb]=np.array([splev(x,tck)])
                    self[limb+"PhaseSig"]=pphase[limb][0]
                    self[limb+"Phase"] = self.phaser.phaserEval(pphase[limb])[0,:]%(2.0*np.pi)
            if self.debug:
                print "limb"+limb
                print "shape of pphase[limb] used to train phase, then of pphase[limb][0]"
		print pphase[limb].shape
		print pphase[limb][0].shape
                print np.any(np.isnan(pphase[limb]))
                print np.sum(np.isnan(pphase[limb]))
                if self.traindata is not None:
                    print "training pphase shapes"
                    for e in trainpphase[limb]:
                        print e.shape
            #print "passed phaser once"
            # Store filtered phase signal
            # Get Phases
        self.relativePhaseLabels = [l1+l2+"RelativePhase" for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])]
        for l1,l2 in zip(self.limbs[:-1],self.limbs[1:]):
            self[l1+l2+"RelativePhase"] = (self[l1+"Phase"]-self[l2+"Phase"])%(2.0*np.pi)

        # Construct a global kinematic phase?
        #self.phaser = Phaser( np.vstack([self[k] for k in self.limbs]), psecfunc=self.laPsec )
        #self["globalPhase"] = self.phaser.phaserEval(np.vstack([self[k] for k in self.limbs]))[0,:]%(2.0*np.pi)
        
    def _strideCut(self):
        # Find strides based on fore left limb phase
        # More things to compute: 
        # - Use angularCentral on striderelatives to compute a "middle" of the cloud
        # - use g3D to compute distances of all poitns to centroid
        # - Compare distribution of distances for various conditions. 
        # -  - compute distance with (point - centroid) dot g3D dot (point - centroid)
        self["relativeStridePhases"]=list()
        self.strideIdx = [idx for e1,e2,idx in zip((self["5Phase"][:-1]+np.pi)%(2.0*np.pi)-np.pi,(self["5Phase"][1:]+np.pi)%(2.0*np.pi)-np.pi,range(len(self["5Phase"][:-1]))) if e1<0 and e2>0]
        # So few strides in spider data, and we're trimming by sample: lets get rid of this.
        #self.firstStrideIdx = self.strideIdx[self.firstStride]
        #self.lastStrideIdx = self.strideIdx[self.lastStride]
        for idx0,idx1 in zip(self.strideIdx[0:-1],self.strideIdx[1:]): # HARDCODED to ignore the first few
            self["relativeStridePhases"].append([
                                                   angularCentralTendancy(self[l1+l2+"RelativePhase"][idx0:idx1],ctf=nanmean)
                                                 for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
        #for idx0,idx1 in zip(self.strideIdx[self.firstStride:self.lastStride-1],self.strideIdx[self.firstStride+1:self.lastStride]): # HARDCODED to ignore the first few
            #self["relativeStridePhases"].append([
            #                                       angularCentralTendancy(self[l1+l2+"RelativePhase"][idx0:idx1],ctf=nanmean)
            #                                     for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
        self["relativeStridePhases"] = np.array(self["relativeStridePhases"])
        self.relativeStridePhaseLabels = [l1+l2+"RelativeStridePhase" for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])]        
        
    def _computeCentroids(self):        
        # Compute centroid of whole waveform relative phases:
        self["relativePhaseCentroid"] = [angularCentralTendancy(self[l],ctf=nanmean) for l in self.relativePhaseLabels]
        # Compute centroid of stride relative phases:
        self["relativeStridePhaseCentroid"] = [angularCentralTendancy(self["relativeStridePhases"][:,k],ctf=nanmean) for k in range(len(self.relativePhaseLabels))]
        self.relativeStridePhaseCentroidLabels = [l1+l2+"RelativeStridePhaseCentroid" for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])]
            
    def plotPhaseComputation(self):
        for limb in self.limbs:
            fig = pylab.figure()
            pylab.clf()
            pylab.plot(zscore(self[limb]))
            pylab.plot(self[limb+"PhaseSig"])
            pylab.plot(self[limb+'Phase'])
            pylab.legend(('raw fore aft','filtered phase sig','phase'))
            pylab.title(limb)
            pylab.savefig(osjoin(self.analysisDirectory,limb+"phaseComputation.pdf"))
            
        pylab.figure(figsize=(13,7.5))
        for j,limb in enumerate(self.limbs):
            ax=pylab.subplot(1,len(self.limbs),j+1)
            wid = 5
            labs = ['leg x pos','phaseSig','phase']
            locs = np.array(range(len(labs)))*5
            mid = np.mean([self.time[0],self.time[-1]])
            pylab.plot(self.time-mid,zscore(self[limb])+locs[0])
            pylab.plot(self.time-mid,zscore(self[limb+"PhaseSig"])+locs[0])
            pylab.plot(self.time-mid,self[limb+'Phase']+locs[1])
            ax.set_yticks(locs)
            ax.set_yticklabels(labs,rotation=90)
            ax.set_xlim([-wid/2,+wid/2])
            pylab.grid()
            if j==0:
                ax.set_ylabel('Signal')
            ax.set_xlabel('Time (sec)')
            ax.set_title(limb)
            ax.set_ylim((-5,5*(len(labs)+1)))
        #ax.annotate( " ".join([self.parent.parent.iD,self.parent["trialType"], \
        #    self.parent["terrainType"],self.parent.iD]),xy=(0.5,1), xytext=(0,-10),\
        #    xycoords='figure fraction',textcoords='offset points',va='top',ha='center')
        #pylab.savefig(osjoin(self.analysisDirectory, "_".join(["phaseComputation", \
        #    self.parent.parent.iD,self.parent["trialType"],self.parent["terrainType"], \
        #    self.parent.iD,".png"])))
        #pylab.close()
        pylab.savefig(osjoin(self.analysisDirectory, "phaseComputation.pdf"))
        pylab.close('all')  
        
    def plotRelativePhases(self):
        sigs = list()
        # this is first limb to second from last zipped with second limb through last:
        #In [19]: zip(imud.limbs[:-1],imud.limbs[1:])
        #Out[19]: [('FL', 'BR'), ('BR', 'BL'), ('BL', 'FR')]
        labs = list()
        for l1,l2 in zip(self.limbs[:-1],self.limbs[1:]):
            sigs.append(self[l1+l2+"RelativePhase"])
            labs.append(l1+l2+"RelPh")
        rph = np.array(sigs).T
        pylab.figure(figsize=(13,7.5))
        #ax1 = pylab.subplot2grid((4,1), (0, 0), rowspan=3)
        pylab.plot(self.time-self.time[0],rph)
        ax1 = pylab.gca()
        pylab.legend(labs)
        pylab.hlines([np.pi/2, np.pi,3*np.pi/2,2*np.pi],0,self.time[-1]-self.time[0])
        [pylab.axvline(x=self.time[k],color='r') for k in self.strideIdx]
        ax1.set_ylabel('Relative leg phase (rad)')
        ax1.set_xticks([])
        #ax1.set_title(" ".join([self.parent.parent.iD,self.parent["trialType"], \
        #    self.parent["terrainType"],self.parent.iD]))
        ax1.set_ylim([0, np.pi*2])
        ax1.set_xlim((0,self.time[-1]-self.time[0]))
        pylab.savefig(osjoin(self.analysisDirectory, "relativePhases.pdf"))
        pylab.savefig(osjoin(self.analysisDirectory, "relativePhases.svg"),transparent=True)
        pylab.close('all')
        #pylab.show()
        #pylab.savefig(osjoin(self.analysisDirectory, "_".join(["relativePhases", \
        #    self.parent.parent.iD,self.parent["trialType"],self.parent["terrainType"], \
        #    self.parent.iD,".png"])))

    def plot3DPhaseScatterPlot(self):
        x=self["relativeStridePhases"].transpose()[0]
        y=self["relativeStridePhases"].transpose()[1]
        z=self["relativeStridePhases"].transpose()[2]
        fig=pylab.figure()
        ax = p3.Axes3D(fig)
        ax.scatter3D(np.ravel(x),np.ravel(y),np.ravel(z),c='r')
        ax.set_xlabel('Phi1')
        ax.set_ylabel('Phi2')
        ax.set_zlabel('Phi3')
        ax.set_xlim3d([0,2.0*np.pi])
        ax.set_ylim3d([0,2.0*np.pi])
        ax.set_zlim3d([0,2.0*np.pi])
        fig.add_axes(ax)
        #fig.legend([(l1+l2+"RelativePhase") for l1,l2 in zip(self.limbs[:-1],self.limbs[1:])])
                     
                    
    def plotRawData(self):
        pylab.figure()
        pylab.clf()
        pylab.plot(self.res.pthX,self.res.pthY)
        pylab.title("Raw leg positions in body coordinates")
        pylab.savefig(osjoin(self.analysisDirectory, "rawLegPosBodyCoords.pdf"))
        
        # Add raw leg pthX with vertbars at trimstart trimstop
        pylab.clf()
        pylab.plot(self.res.pthX)
        [pylab.axvline(x=p,color='k') for p in [self.trimStart,(self.res.pthX.shape[0]+self.trimEnd) if self.trimEnd < 0 else self.trimEnd]]
        pylab.title("Leg fore-aft coordinates (x) blackvbars=trim redvbvar=stride")
        [pylab.axvline(x=k,color='r') for k in self.strideIdx+np.array(self.trimStart)]

        pylab.savefig(osjoin(self.analysisDirectory, "rawLegForeAft.pdf"))
        
        pylab.close('all')
        
        # Add top down animal position and speed -- if loaded from imgs
                
             
    def writeData(self):
        '''
        Write data
        '''
        print "TODO"
    
    def doAnalysis(self):
        self._getPhases()
        self._strideCut()
        #self._computeCentroids()
        self.plotRawData()
        self.plotPhaseComputation()
        self.plotRelativePhases()        
        #self.plot3DPhaseScatterPlot()
    
    def __init__(self, workingDirectory, analysisDirectory, metaData ):
        '''
        Constructor - Runs through the procedure of calculating computationally cheap parameters
        such as angles and lengths
        '''
        DirectoryAware.__init__(self, workingDirectory,analysisDirectory,metaData)
        self.fileName = metaData["fileName"]
        self.sampleRate = metaData["sampleRate"]
        self.data = dict()
        self.strideIdx = list()
        if metaData.has_key("firstStride"):
            self.firstStride = metaData["firstStride"]
            if metaData.has_key("numStrides"):
                self.numStrides = metaData["numStrides"]
                self.lastStride = self.firstStride + self.numStrides
            else:
                self.lastStride = metaData["lastStride"]
        else:
            self.firstStride = 1
            self.lastStride = -1
        if metaData.has_key("trimStart"):
            self.trimStart = metaData["trimStart"]
        else:  
            self.trimStart = 1
        if metaData.has_key("trimEnd"):
            self.trimEnd = metaData["trimEnd"]
        else:  
            self.trimEnd = -1
        if metaData.has_key("trialType"):
            self.trialType=metaData["trialType"]
        else:
            raise("ERROR: No trial type specified for spiderTarsiTrackData.")
        if metaData.has_key("smoothfactor"):
            self.smoothfactor = metaData["smoothfactor"]
        else:  
            # Larger = more detail, smaller = more smooth. 
            self.smoothfactor = 40.0
        if self.trialType == "intact":
            self.limbs=["1","2","3","4","5","6","7","8"]
        elif self.trialType == "ablated":
            self.limbs = ["1","2","3","5","6","7"]
        if metaData.has_key("debug"):
            self.debug=metaData("debug")
        else:
            self.debug=False

        self.traindata=None
        self._readData()
        if self.debug:
            self.plotRawData()
        
def getSampleSpiderTarsiTrackingObject():
    '''This function returns an example of a motion capture dataModule
    object, based on the directory structure of Alexis' data from fish. 
    '''
    workingDir = osjoin("..")    
    analysisDir = osjoin("..")
    metaDataFile = osjoin("..",'spiderGaitAblations','spiderGait.yaml')
    stream = file(metaDataFile, 'r')
    metaData=yamlload(stream)
    stream.close()
    return(SpiderTarsiTrackData(workingDir,analysisDir,metaData))  

if __name__=='__main__':
    '''
    Runs some tests using the sample data
    '''
    getSampleSpiderTarsiTrackingObject()
    
