'''
This version uses the convention of leg phase differences we quote in the paper,
with legs:
0 3
2 1

th1=0-1
th2=1-2
th3=2-3

Lateral walk (which our data contains) is 0.25, 0.5, 0.75

NOTE:
Old phaseDiffs flips sign above above th, such that
th1=1-0
th2=2-1
th3=3-2
'''

from numpy import *
import sys
import sympy as sp

oldhook = sys.displayhook
sys.displayhook = sp.pprint

def makeSinCos(gaits):
    '''
    Compute the matricies for various gaits
    '''
    Ng = gaits.shape[0]
    C1 = array([[sp.cos(2*sp.pi*subgait) for subgait in gait] for gait in gaits])
    S1 = array([[sp.sin(2*sp.pi*subgait) for subgait in gait] for gait in gaits])
    return(C1,S1)

def makeEquations(C1,S1,gamma,epsilon,sigma,omega,eta,rhs):
    return(sp.Matrix(
           sum(gamma*C1,1) +
           sum(epsilon*S1,1) +
           diagonal(dot(S1,dot(sigma,S1.transpose()))) +
           diagonal(dot(C1,dot(omega,C1.transpose()))) +
           diagonal(dot(S1,dot(eta,C1.transpose()))) +
           - array(rhs)
    ))
    
def makeDerivative(C1,S1,gamma,epsilon,sigma,omega,eta):
    return(sp.Matrix(
        -gamma*S1 +
        epsilon*C1 +
        2*dot(S1,sigma)*C1 +
        -2*dot(C1,omega)*S1 +
        dot(C1,eta.transpose())*C1 - dot(S1,eta)*S1
    ))

def makeDDerivative(C1,S1,gamma,epsilon,sigma,omega,eta):
    return(sp.Matrix(
        -gamma*C1 +
        -epsilon*S1 +
        -2*dot(S1,sigma)*S1 +
        -2*dot(C1,omega)*C1 +
        -dot(C1,eta.transpose())*S1 - dot(S1,eta)*C1
    ))

def makeODDDerivative(C1,S1,gamma,epsilon,sigma,omega,eta):
    beta = [2,0,1]
    alpha = [1,2,0]
    return(sp.Matrix(
                      (2*C1[:,beta]*sigma[beta,alpha]-S1[:,beta]*eta.transpose()[beta,alpha])*C1[:,alpha] +
                      (2*S1[:,beta]*omega[beta,alpha]-C1[:,beta]*eta[beta,alpha])*S1[:,alpha]
                     ))

g3T = 1.0/4.0*array([
                     [3,2,1],
                     [2,4,2],
                     [1,2,3]
                    ])


# Gaits and Costs
walkGaits = sp.Rational(1,4)*array([
                                    [2,3,2],
                                    [2,1,2],
                                    [1,1,1],
                                    [3,3,3],
                                    [1,2,3],
                                    [3,2,1]
                                   ])
                                   
walkLongScores = array([
                        0,
                        0,
                        1,
                        1,
                        -1,
                        1
                       ])

pronkGaits = array([
                    [0,0,0]
                   ])

otherStationaryGait = sp.Rational(1,6)*array([
                                              [sp.Rational(3,2),1,5]
                                             ])
                                      
# Parameters
gamma = array([sp.Symbol('g0'),sp.Symbol('g1'),sp.Symbol('g2')])
epsilon = array([sp.Symbol('e0'),sp.Symbol('e1'),sp.Symbol('e2')])
sigma = array([
               [0,sp.Symbol('s10'),sp.Symbol('s20')],
               [sp.Symbol('s10'),0,sp.Symbol('s21')],
               [sp.Symbol('s20'),sp.Symbol('s21'),0]
              ]) 

omega = array([
               [0,sp.Symbol('w10'),sp.Symbol('w20')],
               [sp.Symbol('w10'),0,sp.Symbol('w21')],
               [sp.Symbol('w20'),sp.Symbol('w21'),0]
              ])

eta = array([
             [0,sp.Symbol('eta01'),sp.Symbol('eta02')],
             [sp.Symbol('eta10'),0,sp.Symbol('eta12')],
             [sp.Symbol('eta20'),sp.Symbol('eta21'),0]
            ])

C1,S1 = makeSinCos(walkGaits)

print "Cosine matrix of walk gaits", C1
print "Sine matrix of walk gaits", S1

e = makeEquations(C1,S1,gamma,epsilon,sigma,omega,eta,walkLongScores)
de = makeDerivative(C1,S1,gamma,epsilon,sigma,omega,eta)
dde = makeDDerivative(C1,S1,gamma,epsilon,sigma,omega,eta)
ddde = makeODDDerivative(C1,S1,gamma,epsilon,sigma,omega,eta)

C1,S1 = makeSinCos(pronkGaits)

pe = makeEquations(C1,S1,gamma,epsilon,sigma,omega,eta,[0]) 
pde = makeDerivative(C1,S1,gamma,epsilon,sigma,omega,eta)
pdde = makeDDerivative(C1,S1,gamma,epsilon,sigma,omega,eta)

C1,S1 = makeSinCos(otherStationaryGait)

osge = makeEquations(C1,S1,gamma,epsilon,sigma,omega,eta,[0]) 
osgde = makeDerivative(C1,S1,gamma,epsilon,sigma,omega,eta,)
osgdde = makeDDerivative(C1,S1,gamma,epsilon,sigma,omega,eta) 

def applyTrans(trans,e):
    for s,t in trans:
        e = e.subs(s,t)
    return(e)

def prntLastTrans(trans):
    print str(trans[-1][0]) + " = " + str(trans[-1][1])

# Save the necessary transformations
trans = []

# Solve Stability Equation
print "Starting with this system of equations, from sin cose matrices * params - longstab = 0"
print e

print "Solving the stability equation: e.g. applying transformations in sequence to above equations until 0"
trans.append((epsilon[1],epsilon[1]-sp.Rational(1,2)*(e[2]-e[3])))
prntLastTrans(trans)
e = applyTrans(trans,e)

trans.append((sigma[2,1],sigma[2,1]-sp.Rational(1,2)*e[2]))
prntLastTrans(trans)
e = applyTrans(trans,e)

trans.append((omega[2,0],omega[2,0]-sp.Rational(1,4)*(e[0]+e[1])))
prntLastTrans(trans)
e = applyTrans(trans,e)

trans.append((eta[1,2],eta[1,2]-e[0]))
prntLastTrans(trans)
e = applyTrans(trans,e)

trans.append((eta[1,2],eta[1,2]-e[0]))
prntLastTrans(trans)
e = applyTrans(trans,e)

trans.append((sigma[2,0],sigma[2,0]+sp.Rational(1,4)*(e[4]+e[5])))
prntLastTrans(trans)
e = applyTrans(trans,e)

trans.append((eta[2,1],eta[2,1]-e[4]))
prntLastTrans(trans)
e = applyTrans(trans,e)

# Check
print "System of equations after all of this (should be zero):"
print e # Should be zero

print "Applying all those transformations to the derivative equations"
de = applyTrans(trans,de)
print "Derivative equations now:", de

# Now make the walk gait a stationary
print "Now making the walking gait stationary:"

print "THIS USED TO WORK BY SETTING EXAMINING THE DERIVATIVE EQUATIONS AT THIS POINT"
print "AND SETTING ONE PARAMETER IN THEM IN EACH COMPONENT SUCH THAT IT CANCELLED ALL"
print "THE REST"
print "However, we now want gait 4, not gait 5, to be stationary, so we adjust!"
print "Examining the equations now, we can still use the same paramter to cancel the others"
print "Must just change to row 4"
trans.append((eta[2,0],eta[2,0]-de[4,0]))
prntLastTrans(trans)
de = applyTrans(trans,de)

trans.append((eta[0,2],eta[0,2]-de[4,2]))
prntLastTrans(trans)
de = applyTrans(trans,de)

'''
Hrm, sign opposite:
will get:

stuff - 4 * (s10 - 1/4(stuff-4*s10))
= stuff - 4s10 + stuff -4s10
do it again:
stuff - 4s10 + 1(stuff-4s10)
= stuff-4s10+stuff-4s10 = 2stuff-8s10

if do s10->s10+1/4stuff get

stuff - 4(s10+1/4(stuff-4s10))
= stuff - 4s10 - 1(stuff-4s10)
= stuff -4s10 -stuff + 4s10 = 0
'''
trans.append((sigma[1,0],sigma[1,0]+sp.Rational(1,4)*de[4,1]))
prntLastTrans(trans)
de = applyTrans(trans,de)
print "Now row five of this  should be zero"
print de

"Applything all transformations to dde"
dde = applyTrans(trans,dde)
print "dde is now"
print dde

# Now parameterise this so it will be a minima
z = sp.Symbol('z')
y = sp.Symbol('y')

print "Adding transformations to parametrize a minimum at long stab walk"
print "Examining dde, set eta01 to itself minus dde4,0 minus param=y"
print "Sign flip from earlier potential as eqn now is plus eta01"
trans.append((eta[0,1],eta[0,1]-dde[4,0]+y))
prntLastTrans(trans)
dde = applyTrans(trans,dde)

print "gamma remains negative in dde4,2 as was in dde5,2"
print "Hang on why 1/2? -- ah, so you get y,z,y in row"
trans.append((gamma[1],gamma[1]+sp.Rational(1,2)*(dde[4,2]-y)))
prntLastTrans(trans)
dde = applyTrans(trans,dde)

print "Epsilon0 was negative in previous version: now positive."
trans.append((epsilon[0],epsilon[0]-dde[4,1]+z))
prntLastTrans(trans)
dde = applyTrans(trans,dde)

print "dde now", dde

e = applyTrans(trans,e)
de = applyTrans(trans,de)

pe =  applyTrans(trans,pe)
pde =  applyTrans(trans,pde)

# Make the all maximally unstable walks stationary
print "making all maximally unstable walks stationary"
print "By setting params in de to cancel:"
print "de is now", de
print "want rows 3,4, and 6 (1 based), to zero):"
print "Start with diagonal walk: last row"
print "Killing de5,2 with omega21"
trans.append((omega[2,1],omega[2,1]-sp.Rational(1,4)*de[5,2]))
prntLastTrans(trans)
de = applyTrans(trans,de)
print de

print "killing de51 with epsilon"
trans.append((epsilon[2],epsilon[2]-sp.Rational(1,4)*de[5,1]))
prntLastTrans(trans)
de = applyTrans(trans,de)
print de

print "killing de50 with omega10"
trans.append((omega[1,0],omega[1,0]+sp.Rational(1,4)*de[5,0]))
prntLastTrans(trans)
de = applyTrans(trans,de)
print de

print "killing de3,1 with y"
trans.append((y,y+de[3,1]))
prntLastTrans(trans)
de = applyTrans(trans,de)
print de

print "killing de32 with gamma2"
trans.append((gamma[2],gamma[2]-de[3,2]))
prntLastTrans(trans)
de = applyTrans(trans,de)
print de

print "killing de30 with eta10"
trans.append((eta[1,0],eta[1,0]+de[3,0]))
prntLastTrans(trans)
de = applyTrans(trans,de)

print "killing de20 with gamma0"
trans.append((gamma[0],gamma[0]+sp.Rational(1,2)*de[2,0]))
prntLastTrans(trans)
de = applyTrans(trans,de)

print de

# Apply transformations
pe =  applyTrans(trans,pe)
pde =  applyTrans(trans,pde)
de =  applyTrans(trans,de)
dde =  applyTrans(trans,dde)
osge = applyTrans(trans,osge)
osgde = applyTrans(trans,osgde)
osgdde = applyTrans(trans,osgdde)
ddde = applyTrans(trans,ddde)

Cs = array([[sp.Symbol('cos(x[0])'),sp.Symbol('cos(x[1])'),sp.Symbol('cos(x[2])')]])
Ss = array([[sp.Symbol('sin(x[0])'),sp.Symbol('sin(x[1])'),sp.Symbol('sin(x[2])')]])

# Calculate Potential
V = makeEquations(Cs,Ss,gamma,epsilon,sigma,omega,eta,[0])[0]
dV = makeDerivative(Cs,Ss,gamma,epsilon,sigma,omega,eta)
ddV = makeDDerivative(Cs,Ss,gamma,epsilon,sigma,omega,eta)

V =  applyTrans(trans,V)
dV =  applyTrans(trans,dV)

sys.displayhook = oldhook
V = sp.simplify(V)
dV.simplify()
ddV.simplify
print "\nV"
print V
print "\ndV"
print dV
print "\nddV"
print ddV

'''
V
cos(x[1])*sin(x[0])*z/2 - cos(x[1])*sin(x[2])*z/2 + sin(x[0])*sin(x[1])/2 + sin(x[0])*z/2 - sin(x[0])/2 + sin(x[1])*sin(x[2])/2 - sin(x[2])*z/2 + sin(x[2])/2

dV
[cos(x[0])*(cos(x[1])*z + sin(x[1]) + z - 1)/2,

cos(x[1])*sin(x[0])/2 + cos(x[1])*sin(x[2])/2 - sin(x[0])*sin(x[1])*z/2 + sin(x[1])*sin(x[2])*z/2,

cos(x[2])*(-cos(x[1])*z + sin(x[1]) - z + 1)/2]
'''

def Potential(x,z):
    return(
        cos(x[1])*sin(x[0])*z/2 -
        cos(x[1])*sin(x[2])*z/2 +
        sin(x[0])*sin(x[1])/2 +
        sin(x[0])*z/2 -
        sin(x[0])/2 +
        sin(x[1])*sin(x[2])/2 -
        sin(x[2])*z/2 +
        sin(x[2])/2
       )

def dPotential(x,z):
    return(array(
            [
             cos(x[0])*(
                        cos(x[1])*z +
                        sin(x[1]) +
                        z -
                        1
                       )/2,
             cos(x[1])*sin(x[0])/2 +
             cos(x[1])*sin(x[2])/2 -
             sin(x[0])*sin(x[1])*z/2 +
             sin(x[1])*sin(x[2])*z/2,
             cos(x[2])*(
                        -cos(x[1])*z +
                        sin(x[1]) -
                        z +
                        1
                       )/2
            ]))

walkGaits = 1.0/4.0*array([
                                    [2,3,2],
                                    [2,1,2],
                                    [1,1,1],
                                    [3,3,3],
                                    [1,2,3],
                                    [3,2,1]
                                   ])    

print "WalkGaits", walkGaits
# Check scores at z=0
print "Potential evaluated at walkGaits: z=0"
print [Potential(2.0*pi*w,0) for w in walkGaits]
print "Potential at gaits minus longstab scores < 1e-10, z=0"
print abs(array([Potential(2.0*pi*w,0) for w in walkGaits]) - walkLongScores) < 1e-10

# Check scores at z=0.5
print "Same for z=0.5"
print [Potential(2.0*pi*w,0.5) for w in walkGaits]
print abs(array([Potential(2.0*pi*w,0.5) for w in walkGaits]) - walkLongScores) < 1e-10

# Check scores
print "Same for z=1.0"
print [Potential(2.0*pi*w,1.0) for w in walkGaits]
print abs(array([Potential(2.0*pi*w,1.0) for w in walkGaits]) - walkLongScores) < 1e-10

# Check Gradients
print "Gradient, dPotential, at walkgaits, and then err < tol, z=0"
print [dPotential(2.0*pi*w,0) for w in walkGaits]
print abs(array([dPotential(2.0*pi*w,0) for w in walkGaits])) < 1e-10

# Check Gradients
print "Gradient, dPotential, at walkgaits, and then err < tol, z=1.0"
print [dPotential(2.0*pi*w,1.0) for w in walkGaits]
print abs(array([dPotential(2.0*pi*w,1.0) for w in walkGaits])) < 1e-10     

#Check Gradients match up
print "Gradient wrt phi0 component in +phi0 dir at z=1 for gait 4 (lateral walk now, most stable)"
print "Should be positive: small dev minus value at minimum" 
print array([dPotential(2.0*pi*w+[0.001,0.0,0.0],1.0)-dPotential(2.0*pi*w,1.0) for w in walkGaits])[4,0]/0.001
print "Gradient wrt phi1 component in +phi0 dir at z=1 for gait 4 (lateral walk now, most stable)" 
print "Should be positive: small dev minus value at minimum" 
print array([dPotential(2.0*pi*w+[0.0,0.001,0.0],1.0)-dPotential(2.0*pi*w,1.0) for w in walkGaits])[4,1]/0.001
print "Gradient wrt phi2 component in +phi0 dir at z=1 for gait 4 (lateral walk now, most stable)" 
print "Should be positive: small dev minus value at minimum" 
print array([dPotential(2.0*pi*w+[0.0,0.0,0.001],1.0)-dPotential(2.0*pi*w,1.0) for w in walkGaits])[4,2]/0.001

# Now show parameters
p = []
p.extend(sp.Matrix(gamma).mat)
p.extend(sp.Matrix(epsilon).mat)
p.extend(sp.Matrix(sigma).mat)
p.extend(sp.Matrix(omega).mat)
p.extend(sp.Matrix(eta).mat)
p = sp.Matrix(p)
op = p
p = applyTrans(trans,p)
for pi,qi in zip(p,op):
    print qi, " = ", pi
