'''
Created on 18 Feb 2011

@author: Andrew Spence, Simon Wilshin, 
         Some code adapted from contributions by Shai Revzen
'''
import numpy as np
import pylab
from matplotlib.patches import Polygon

def insetAxes(axs,ticklabels=True):
    '''
    Returns a new axes in the upper left corner of the given axes
    ticklabels determines whether tick labels are visible.
    '''
    rct=np.array(axs.get_position())
    newax=pylab.axes([(rct[0,0]+0.1*(rct[1,0]-rct[0,0])), (rct[0,1]+0.1*(rct[1,1]-rct[0,1])), (0.25*(rct[1,0]-rct[0,0])), (0.25*(rct[1,1]-rct[0,1]))])
    if not ticklabels:
        pylab.setp(newax.get_xticklabels(),visible=False)
        pylab.setp(newax.get_yticklabels(),visible=False)

def vertbars(x,**args):
    '''Plots vertical bars that span the current axis y limits at the given
    array of x coordinates. Optional keyword arguments are passed to plot.'''
    yl = pylab.ylim()
    return(pylab.plot(np.tile(x,(2,1)),np.tile(np.array(yl),(x.shape[0],1)).T, **args))

def horzbars(y,**args):
    '''Plots vertical bars that span the current axis y limits at the given
    array of x coordinates. Optional keyword arguments are passed to plot.'''
    xl = pylab.xlim()
    #print np.tile(np.array(xl),(y.shape[0],1)).T
    #print np.tile(y,(2,1))
    return(pylab.plot(np.tile(np.array(xl),(y.shape[0],1)).T,np.tile(y,(2,1)), **args))

# Nice patch plotting function:
def plotPatch(x,upper,lower,inalpha=0.2,fcolor='red',inlw=0):
    '''
    Plots up a shaded patch covering the polygon
    defined by the lines defined by the x, uppwer, and lower coordinates
    given.
    Expects x, upper, lower as equal length 1D arrays.
    x, upper, and lower are expected in increasing order.
    '''
    verts = zip( x, upper )
    verts.extend( zip( x[::-1], lower[::-1] ) )
    mypatch = Polygon(verts,facecolor=fcolor,edgecolor=fcolor,alpha=inalpha,lw=inlw )
    pylab.gca().add_patch(mypatch)
   
def plotPatchWithLine(x,y,upper,lower,inalpha=0.2,fcolor='red',patchlw=0,linew=1):
    plotPatch(x,upper,lower,inalpha,fcolor,patchlw)
    pylab.gca().plot(x,y,lw=linew,color=fcolor)

def plotPatchRadial(x,upper,lower,inalpha=0.2,fcolor='red',inlw=0):
    '''
    Plots up a shaded patch covering the polygon
    defined by the lines defined by the x, uppwer, and lower coordinates
given.
    Expects x, upper, lower as equal length 1D arrays.
    x, upper, and lower are expected in increasing order.
    '''
    verts = zip( upper, x )
    verts.extend( zip( lower[::-1], x[::-1] ) )
    mypatch = Polygon(verts,facecolor=fcolor,edgecolor=fcolor,alpha=inalpha,lw=inlw )
    pylab.gca().add_patch(mypatch)

def plotPatchWithLineRadial(x,y,upper,lower,inalpha=0.2,fcolor='red',patchlw=0,linew=1):
    plotPatchRadial(x,upper,lower,inalpha,fcolor,patchlw)
    pylab.gca().plot(y,x,lw=linew,color=fcolor)
    
def testPatches():
    # Test out me functions:
    x = np.arange(0.,1.,0.001)
    #y = np.sin(2.0*np.pi*x) + 0.1 * np.random.standard_normal(len(x)) 
    y = np.sin(2.0*np.pi*x)
    
    upper = y + 0.1
    lower = y - 0.1
    
    pylab.figure()
    plotPatch(x,upper,lower)
    pylab.figure()
    plotPatchWithLine(x,y,upper,lower,0.3,'blue',linew=2)
    pylab.figure()
    plotPatchWithLine(x,y,upper,lower,0.3,'blue',patchlw=1,linew=2)
    
    
class selUI( object ):
    """
    Adapted from Shai Revzen, U Penn, 2010 python_libs package
    by Simon Wilshin Royal Veterinary College, 2011.  
    Allow user to click to highlight and identify line handles
    """
    def __init__(self, lines=None):
        if lines is None:
            lines = pylab.gca().get_lines()
        assert all([ isinstance( h, pylab.Line2D ) for h in lines ])
        self.lh = lines
        self.fig = pylab.gcf()
        self.ax = pylab.gca()
        self.o_lw = []
        self.o_lal = []
        self.fig.set_visible(False)
        for lh in lines:
            self.o_lw.append( lh.get_lw() )
            self.o_lal.append( lh.get_alpha() )
            lh.set( lw = 1, alpha = 0.4, picker=2 )
        self.sel = None
        self.lbl = None
        self.fig.canvas.mpl_connect( 'pick_event', self._doPick )
        self.fig.set_visible(True)
        pylab.draw()
  
    def _lblUpdate( self, x, y, N ):
        if self.lbl is None:
            lbl = pylab.text( x, y, str(N) )
            lbl.set_backgroundcolor((1,1,1))
            self.lbl = lbl
        else:
            lbl = self.lbl
        lbl.set( x = x, y = y, text = str(N), visible=True )
      
    def _axesPicked( self ):
        if self.sel is not None:
            l = self.lh[self.sel]
            l.set( lw = 1, alpha = 0.4 )
            self.sel = None
        if self.lbl is not None:
            self.lbl.set_visible(False)
    
    def _linePicked( self, sel, x, y ):
        if self.sel is not None:
            l = self.lh[self.sel]
            l.set( lw = 1, alpha = 0.2 )
            self.sel = None
        self.sel = sel
        self.lh[sel].set( 
                         alpha = self.o_lal[ self.sel ], 
                         lw = self.o_lw[ self.sel ] + 1 
                        )
        if len(str(self.lh[sel].get_label()))<5 or str(self.lh[sel].get_label())[:5]=='_line':
            self._lblUpdate( x,y, sel )
        else:      
            self._lblUpdate( x,y, self.lh[sel].get_label())
            
    def _doPick( self, event ):
        try:
            idx = self.lh.index( event.artist )
            self._linePicked( idx, event.mouseevent.xdata, event.mouseevent.ydata )
        except ValueError:
            self._axesPicked()
        pylab.draw()

