
from numpy import *
from kalman import Kalman
from scipy.interpolate import interp1d
import pylab as pl
from phaseShifter import *

       
class VXPhase(object):
    '''
    Callable which computes the protophase in position-velocity  space for some signal.
    Requires the sample frequency and the amount to scale position so that it has the same magnitude as the 
    computed velocity, and whether or not to flip the protophase if motion runs 
    clockwise rather than  counterclockwise in the x-v plane.
    '''
    def __init__(self, sampFreq = 1.0, yscale = 1.0, flip = False):
        '''
        sampFreq: Sample frequency (used to scale derivatives)
        yscale: Amount to scale positions so that v-x plots look like circles
        flip: If true then the phase will run in the opposite direction to normal,
              use if phase is monotonically decreasing rather than increasing.
        '''
        self.yscale = yscale
        if flip:
            self.flip = -1.0
        else:
            self.flip = 1.0
        self.sampFreq = sampFreq
    
    def _numericallyDiff(self,dat):
        # Numerically differentiate and interpolate back to original spacing
        v_dat = (dat[:,1:]-dat[:,:-1])*self.sampFreq
        oRange = array(range(dat.shape[1]))-1.0
        nRange = array(range(v_dat.shape[1]))
        oRange[0] += 1
        return(interp1d(nRange,v_dat)(oRange))
        
    def __call__(self,dat):
        v_dat = self._numericallyDiff(dat)
        # Return equivalent of analytic signal
        return v_dat+(dat*1.0j)*self.yscale*self.flip

class VXKalmanPhase(VXPhase):
    '''
    Construct a similar phase to VXPhase but uses a very 
    crude Kalman filter to perform smoothing.
    '''
    def __init__(self, sampFreq = 1.0, yscale = 1.0, flip = False):
        VXPhase.__init__(self, sampFreq, yscale, flip)
    
    @reCenterCl
    def __call__(self,dat):
        v_dat = self._numericallyDiff(dat)
        
        # Run a Kalman filter in reverse to get initial conditions
        k=[Kalman(2,[0,0],A=array([[1.0,-1.0/self.sampFreq],[0,1]])) for d,v in zip(dat,v_dat)]
        for d,v,kf in zip(dat,v_dat,k):
            for d0,v0 in zip(d[::-1],v[::-1]):
                kf.update([d0,v0])
        s = [kf.pmu_hats[-1] for kf in k]
        
        # Make a Kalman filter and smooth the data
        # s = [[d[0],v[0]] for v,d in zip(dat,v_dat)]
        for d,v in s:
            if isnan(d):
                d = 0.0
            if isnan(v):
                v = d/self.sampFreq
            if v==0.0:
                v = 1.0/self.sampFreq
        k=[Kalman(2,[d,v],A=array([[1.0,1.0/self.sampFreq],[0,1]])) for d,v in s]
        for d,v,kf in zip(dat,v_dat,k):
            for d0,v0 in zip(d,v):
                kf.update([d0,v0])
        v_dat = array([array(kf.pmu_hats)[1:,1] for kf in k])
        ndat = array([array(kf.pmu_hats)[1:,0] for kf in k])
        return v_dat+(ndat*1.0j)*self.yscale*self.flip

class VXKalmanPhasePSec(VXKalmanPhase):
    def __init__(self, sampFreq = 1.0, yscale = 1.0, flip = False):
        VXKalmanPhase.__init__(self, sampFreq, yscale, flip)
        
    def __call__(self,dat):
        ph = VXKalmanPhase.__call__(self,dat)
        return(ph.real[0])        
        
def test_sincos():
    """
    Simple test/demo of Phaser, recovering a sine and cosine
    
    Demo courtesy of Jimmy Sastra, U. Penn 2011
    """
    from numpy import sin,cos,pi,array,linspace,cumsum,asarray,dot,ones
    from pylab import plot, legend, axis, show, randint, randn, std,lstsq
    from phaser import Phaser
    # create separate trials and store times and data
    dats=[]
    t0 = []
    period = 55 # i
    phaseNoise = 0.5/sqrt(period)
    snr = 20
    N = 10
    print N,"trials with:"
    print "\tperiod %.2g"%period,"(samples)\n\tSNR %.2g"%snr,"\n\tphase noise %.2g"%phaseNoise,"(radian/cycle)"
    print "\tlength = [",
    for li in xrange(N):
        l = randint(400,2000) # length of trial
        dt = pi*2.0/period + randn(l)*0.07 # create noisy time steps
        t = cumsum(dt)+randn()*2*pi # starting phase is random
        raw = asarray([sin(t),cos(t)]) # signal
        raw = raw + randn(*raw.shape)/snr # SNR=20 noise
        t0.append(t)
        dats.append( raw - raw.mean(axis=1)[:,newaxis] )
        print l,
    print "]"
 
    pPhi = VXKalmanPhase(yscale=2.0*pi/(float(period)),flip=True)
    phr = Phaser( dats, psecfunc = lambda x : dot([1,-1],x),protophfun=pPhi) 
    phi = [ phr.phaserEval( d ) for d in dats ] # extract phaseNoise
    reg = array([linspace(0,1,t0[0].size),ones(t0[0].size)]).T
    tt = dot( reg, lstsq(reg,t0[0])[0] )
    plot(((tt-pi/4) % (2*pi))/pi-1, dats[0].T,'.')
    plot( (phi[0].T % (2*pi))/pi-1, dats[0].T,'x')#plot data versus phase
  
    legend(['sin(t)','cos(t)','sin(phi)','cos(phi)'])
    axis([-1,1,-1.2,1.2])
    show()
    
if __name__=="__main__":
    test_sincos()